#!/usr/bin/env python3
"""
MCP Science Servers Launcher
"""

import sys
import subprocess
import click
from pathlib import Path

# First check if we're in development mode (servers are at project root level)
DEV_SERVERS_PATH = (Path(__file__).parent / '../../servers').resolve()
# Then check if we're in installed mode (servers would be in the package)
INSTALLED_SERVERS_PATH = (Path(__file__).parent / 'servers').resolve()

# Use the development path if it exists, otherwise use the installed path
SERVERS_PATH = DEV_SERVERS_PATH if DEV_SERVERS_PATH.exists() else INSTALLED_SERVERS_PATH
available_servers = [d.name for d in SERVERS_PATH.iterdir() if d.is_dir()] if SERVERS_PATH.exists() else []

@click.command(help="Launch an MCP server by name, installing optional dependencies first.\n\nAvailable servers: {}".format(', '.join(available_servers)))
@click.option('-b', '--branch', type=str, default=None, help='Branch to use for the MCP server', required=False)
@click.argument('server_name', type=str)
@click.argument('args', nargs=-1)
def main(server_name: str, branch: str | None = None, args: list[str] = []) -> None:
    """Launch an MCP server by name, installing optional dependencies first."""
    if branch is None:
        if server_name not in available_servers:
            raise click.BadParameter(
                f"server_name must be one of {available_servers} when --branch is not provided."
            )
        uvx_cmd = [
            "uvx",
            "--from", (SERVERS_PATH / server_name).resolve().as_posix(),
            f"mcp-{server_name}",
            *args,
        ]

    else:
        uvx_cmd = [
            "uvx",
            "--from", f"git+https://github.com/pathintegral-institute/mcp.science@{branch}#subdirectory=servers/{server_name}",
            f"mcp-{server_name}",
            *args,
        ]

    print(f"Running command: {uvx_cmd}")

    try:
        # Run the uvx command
        result = subprocess.run(uvx_cmd, check=False)
        sys.exit(result.returncode)

    except FileNotFoundError:
        print("Error: uvx command not found. Please install uv first.")
        sys.exit(1)
    except Exception as e:
        print(f"Error running {server_name}: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()


from __future__ import annotations

import json
import os
import uuid
from typing import TypedDict

from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from .rester import mp_rester


class SupercellParameters(TypedDict):
    """Parameters for creating and modifying a supercell structure.

    Attributes:
        supercell_size: Size multipliers for each dimension a, b, and c of the unit cell
        removed_region: Region to remove from the structure, specified as [zmin, zmax] where zmin and zmax are relative coordinates between 0 and 1.
    """

    supercell_size: list[int]  # Size multipliers for each dimension [a, b, c]
    removed_region: list[float]  # Region to remove [zmin, zmax]


class MoireParameters(TypedDict):
    pass


class StructureData:
    structure: Structure  # the structure in Pymatgen format
    structure_id: str  # a unique identifier
    material_id: str | None = None  # the material id from MPR in case of bulk structure
    parameters: SupercellParameters | MoireParameters | None = (
        None  # parameters that are used to build the supercell or moire structure
    )
    # parent structures which are usually bulk structures
    parent_ids: list[str] | None = None

    def __init__(
        self,
        structure: Structure
        | str
        | None = None,  # if the structure is a string, it is treated as poscar and converted to Structure
        structure_id: str | None = None,
        material_id: str | None = None,
        parameters: SupercellParameters | MoireParameters | None = None,
        parent_ids: list[str] | None = None,
    ):
        if structure and isinstance(structure, Structure):
            self.structure = structure
        elif structure and isinstance(structure, str):
            self.structure = Structure.from_str(structure, fmt="poscar")
        elif material_id:
            self.material_id = material_id
            result = mp_rester.summary.get_data_by_id(material_id)
            self.structure = result.structure
        else:  # raise error if no structure or material_id
            raise ValueError(
                "Either structure or material_id must be provided")
        if structure_id:
            self.structure_id = structure_id
        else:  # if no structure id is provided, generate a new one
            self.structure_id = str(uuid.uuid4())
        if parameters:
            self.parameters = parameters
        if parent_ids:
            self.parent_ids = parent_ids

    def to_folder(self, folder_path: str):
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        data = {k: v for k, v in self.__dict__.items() if k != "structure"}
        dict_path = os.path.join(folder_path, "structure_info.json")
        with open(dict_path, "w") as f:
            json.dump(data, f, indent=2)
        file_path = os.path.join(folder_path, "structure.cif")
        self.structure.to(fmt="cif", filename=file_path)

    @classmethod
    def from_folder(cls, folder_path: str):
        dict_path = os.path.join(folder_path, "structure_info.json")
        if not os.path.exists(dict_path):
            raise FileNotFoundError(
                f"structure_info.json not found in {folder_path}")

        with open(dict_path, "r") as f:
            data = json.load(f)

        instance = cls.__new__(cls)
        instance.__dict__.update(data)

        file_path = os.path.join(folder_path, "structure.cif")
        if not os.path.exists(file_path):
            raise FileNotFoundError(
                f"structure.cif not found in {folder_path}")
        instance.structure = Structure.from_file(file_path)

        return instance

    @property
    def poscar_str(self) -> str:
        # Create a Poscar object from the structure
        poscar: Poscar = Poscar(structure=self.structure)
        # Convert to a string
        poscar_string: str = poscar.get_str()
        return poscar_string

    @property
    def description(self):
        return self.get_description()

    def get_description(self) -> str:
        description = "Bulk Structure Information\n\n"

        description += f"""
Material id: {self.material_id if 'material_id' in self.__dict__ else 'N/A'}

Species:
{self.structure.species}

Lattice Parameters:
a={self.structure.lattice.a:.4f}
b={self.structure.lattice.b:.4f}
c={self.structure.lattice.c:.4f}
Angles:
alpha={self.structure.lattice.alpha:.4f}
beta={self.structure.lattice.beta:.4f}
gamma={self.structure.lattice.gamma:.4f}

Number of atoms: {len(self.structure)}
"""
        return json.loads(json.dumps(description))

    @classmethod
    def build_supercell(cls, bulk_structure_data: StructureData, supercell_parameters: SupercellParameters):
        bulk_structure = bulk_structure_data.structure
        # Convert to conventional cell
        spacegroup = SpacegroupAnalyzer(bulk_structure)
        conventional_structure = spacegroup.get_conventional_standard_structure()
        # logger.info(conventional_structure)
        supercell_size = [int(n)
                          for n in supercell_parameters["supercell_size"]]

        supercell_structure = conventional_structure * supercell_size

        zs = [float(z) for z in supercell_parameters["removed_region"]]
        # Remove atoms with z > 1-zc or z < zc
        sites_to_remove = []
        for site in supercell_structure:
            if site.frac_coords[2] > zs[0] and site.frac_coords[2] < zs[1]:
                sites_to_remove.append(site)

        if sites_to_remove:
            for site in sites_to_remove:
                supercell_structure.remove(site)

        instance = cls.__new__(cls)
        instance.structure = supercell_structure
        instance.structure_id = str(uuid.uuid4())
        instance.parameters = supercell_parameters
        instance.parent_ids = [bulk_structure_data.structure_id]
        instance.material_id = None
        return instance



import argparse
import json
from ..gpaw_utils.data_class import CalculationProject


def main(project_folder_path: str):
    """
    Create a new project in the given project_folder_path
    """
    calculation_project = CalculationProject(
        project_folder_path=project_folder_path)

    calculation_history = calculation_project.get_calculation_history()
    print(json.dumps({"calculation_history": calculation_history}, indent=4))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--project_folder_path", "-p", type=str)

    args = parser.parse_args()
    main(
        project_folder_path=args.project_folder_path
    )



import logging
import json
from typing import Annotated, cast, Literal
from mcp.server import FastMCP
from pydantic import Field
from mcp.types import ImageContent, TextContent
from jupyter_act.schema import DirectoryContent, FileContent, NotebookContent, Notebook, CodeCell, MarkdownCell, RawCell, CellMetadata
from jupyter_act.utils import (
    add_new_cell_to_notebook,
    execute_code_with_client,
    get_kernel_by_notebook_path,
    get_content,
    replace_cell_in_notebook,
    locate_idx_for_cell_with_given_id
)
import uuid

mcp = FastMCP(name="jupyter", description="Jupyter MCP Server for Code Execution")
logger = logging.getLogger(__name__)

@mcp.tool(
    name="execute_code",
    description="Execute code in a notebook with specific notebook path"
)
def execute_code(
    notebook_path: Annotated[str, Field(..., description="A relative path to the notebook file")],
    code: Annotated[str, Field(..., description="Code to execute")],
) -> TextContent:
    """ Execute code in a notebook with specific notebook path """
    kernel_id = get_kernel_by_notebook_path(notebook_path)
    if kernel_id is None:
        return TextContent(
                type="text",
                text=f"Can not find kernel id with notebook path: {notebook_path}, please make sure the session is running on this notebook"
            )

    # use context manager to manage client resources
    code_cell, run_timeout = execute_code_with_client(kernel_id, code)

    if run_timeout:
        return TextContent(
                type="text",
                text="Waiting execution result timed out. The kernel may still be running the code."
            )

    return TextContent(
            type="text",
            text=code_cell.model_dump_json()
        )


@mcp.tool(
    name="add_cell_and_execute_code",
    description="Add a code cell to the notebook and execute it"
)
def add_cell_and_execute_code(
    notebook_path: Annotated[str, Field(..., description="A relative path to the notebook file")],
    code: Annotated[str, Field(..., description="Code to execute")],
) -> list[TextContent]:
    """ Add a cell to the notebook and execute it """
    kernel_id = get_kernel_by_notebook_path(notebook_path)
    if kernel_id is None:
        return [
            TextContent(
                type="text",
                text=f"Can not find kernel id with notebook path: {notebook_path}, please make sure the session is running on this notebook"
            )
        ]

    # use context manager to manage client resources
    code_cell, run_timeout = execute_code_with_client(kernel_id, code)

    if run_timeout:
        return [
            TextContent(
                type="text",
                text="Waiting execution result timed out. The kernel may still be running the code."
            )
        ]

    # add cell to notebook with put method
    try:
        add_new_cell_to_notebook(notebook_path, code_cell)
    except Exception as e:
        return [
            TextContent(
                type="text",
                text=f"Failed to add cell to notebook: {e}"
            )
        ]

    return [
        TextContent(
            type="text",
            text=code_cell.model_dump_json()
        )
    ]

@mcp.tool(
    name="add_cell",
    description="""
    Add a cell to the notebook. you can add a cell to the notebook with type: code, markdown and raw.
    """
)
def add_cell(
    cell_type: Annotated[Literal["code", "markdown", "raw"], Field(..., description="Cell type, should be 'code' or 'markdown' or 'raw'")],
    notebook_path: Annotated[str, Field(..., description="A relative path to the notebook file")],
    cell_content: Annotated[str, Field(..., description="the content to add to the cell")],
) -> TextContent:
    """ Add a cell to the notebook """
    # Create a new cell based on cell_type
    if cell_type == "code":
        cell = CodeCell(
            id=str(uuid.uuid4()),
            cell_type="code",
            metadata=CellMetadata(),
            source=cell_content,
            outputs=[],
            execution_count=None
        )
    elif cell_type == "markdown":
        cell = MarkdownCell(
            id=str(uuid.uuid4()),
            cell_type="markdown",
            metadata=CellMetadata(),
            source=cell_content
        )
    elif cell_type == "raw":
        cell = RawCell(
            id=str(uuid.uuid4()),
            cell_type="raw",
            metadata=CellMetadata(),
            source=cell_content
        )
    else:
        pass

    try:
        add_new_cell_to_notebook(notebook_path, cell)
    except Exception as e:
        return TextContent(
                type="text",
                text=f"Failed to add cell to notebook: {e}"
            )

    return TextContent(
            type="text",
            text="Cell added successfully"
        )

@mcp.tool(
    name="replace_cell",
    description="Replace the given cell in the notebook. WARNING: This is a high-risk operation that will modify the notebook content without user confirmation. Use with extreme caution and only when absolutely necessary. Ensure you have verified the cell ID and content before proceeding, as this action cannot be undone and may result in data loss if used incorrectly."
)
def replace_cell(
    notebook_path: Annotated[str, Field(..., description="A relative path to the notebook file")],
    cell_id_to_replace: Annotated[str, Field(..., description="The id of the cell to replace")],
    cell_content: Annotated[str, Field(..., description="The Cell Content to replace")],
) -> TextContent:

    # create a code cell with the given content
    cell = CodeCell(
        id=str(uuid.uuid4()),
        cell_type="code",
        metadata=CellMetadata(),
        source=cell_content,
        outputs=[],
        execution_count=None
    )
    # probably we should first validate that there do exists a cell with the given cell id
    try:
        notebook_content = get_content(notebook_path)
    except Exception as e:
        return TextContent(
                type="text",
                text=f"Failed to get notebook content: {e}"
            )

    notebook = cast(Notebook, notebook_content.root.content)
    cell_idx = locate_idx_for_cell_with_given_id(notebook, cell_id_to_replace)
    if cell_idx is None:
        return TextContent(
                type="text",
                text=f"Can not find cell with id: {cell_id_to_replace}"
            )
    # replace cell
    try:
        replace_cell_in_notebook(notebook_path, notebook, cell, cell_idx)
    except Exception as e:
        return TextContent(
                type="text",
                text=f"Failed to replace cell: {e}"
            )

    return TextContent(
            type="text",
            text="Cell replaced successfully"
        )


@mcp.tool(
    name="read_jupyter_file",
    description="Read file content. For regular files, returns the entire content; for .ipynb files, returns notebook cells in pages. Each read returns metadata containing cursor, has_more, and reverse values for subsequent reads. For example, if returned metadata is {\"cursor\": 10, \"has_more\": true}, the next read should use {\"cursor\": 10} as parameter."
)
def read_content(
    file_path: Annotated[str, Field(..., description="File path, supports both regular files and .ipynb files, this path should be the relative path to the working directory")],
    cursor: Annotated[int, Field(default=0, description="Starting position for reading, only applies to .ipynb files. Use 0 for first read, then use the cursor value from previous read's metadata")],
    cell_count: Annotated[int, Field(default=10, description="Number of cells to read, only applies to .ipynb files")],
    reverse: Annotated[bool, Field(default=True, description="Reading direction, only applies to .ipynb files. True means read from end to beginning, False means read from beginning to end. Should match the reverse value from previous read's metadata")],
) -> list[TextContent | ImageContent]:
    try:
        # get file content
        content_response = get_content(file_path)

        # handle different file types
        if content_response.root.type == "file":
            return _handle_file_content(cast(FileContent, content_response.root))
        elif content_response.root.type == "notebook":
            return _handle_notebook_content(cast(NotebookContent, content_response.root), cursor, cell_count, reverse)
        else:
            return [
                TextContent(
                    type="text",
                    text=f"Unsupported file type: {content_response.root.type}"
                )
            ]

    except Exception as e:
        return [
            TextContent(
                type="text",
                text=f"Error reading file: {str(e)}"
            )
        ]

def _handle_file_content(file_content: FileContent) -> list[TextContent | ImageContent]:
    """handle file content"""
    # check if it is an image file
    mimetype = file_content.mimetype
    if mimetype and mimetype.startswith("image/"):
        # for image files, return the content directly
        if file_content.content:
            return [
                ImageContent(
                    type="image",
                    mimeType=mimetype,
                    data=file_content.content
                )
            ]
        else:
            return [
                TextContent(
                    type="text",
                    text="Image content is empty or not available"
                )
            ]

    if file_content.content is None:
        return [
            TextContent(
                type="text",
                text="File content is empty or not available"
            )
        ]

    return [
        TextContent(
            type="text",
            text=file_content.content
        )
    ]

def _handle_notebook_content(notebook_content: NotebookContent, cursor: int, cell_count: int, reverse: bool) -> list[TextContent | ImageContent]:
    """handle notebook content"""
    # get notebook cells
    if notebook_content.content is None or not notebook_content.content.cells:
        return [
            TextContent(
                type="text",
                text="Notebook content is empty or not available"
            )
        ]

    cells = notebook_content.content.cells
    total_cells = len(cells)

    # calculate the cell range to return
    if reverse:
        # Ensure we don't go beyond available cells
        actual_cursor = min(cursor, total_cells)
        start = max(0, total_cells - actual_cursor - cell_count)
        end = total_cells - actual_cursor
        selected_cells = cells[start:end]
        new_cursor = actual_cursor + (end - start)  # Update cursor by the actual number of cells read
        has_more = start > 0  # More content if we haven't reached the beginning
    else:
        start = min(cursor, total_cells)
        end = min(cursor + cell_count, total_cells)
        selected_cells = cells[start:end]
        new_cursor = end  # Next cursor is the end position
        has_more = new_cursor < total_cells  # More content if we haven't reached the end

    # handle each cell content
    result = []

    for i, cell in enumerate(selected_cells):
        # compatible with notebook without cell.id
        global_cursor = start + i

        source = cell.source
        if isinstance(source, list):
            source = "\n".join(source)

        result.append(
            TextContent(
                type="text",
                text=json.dumps( # type: ignore
                    {
                        "cell_type": cell.cell_type,
                        "cell_id": cell.id or global_cursor,
                        "content_type": "source",
                        "content": source
                    }
                )
            )
        )

        if cell.cell_type == "code" and hasattr(cell, "outputs"):
            for output_idx, output in enumerate(cell.outputs):
                if output.output_type in ["execute_result", "display_data"]:
                    for mime_type, data in output.data.root.items(): # type: ignore
                        if mime_type.startswith("image/"):
                            # return image data
                            result.append(
                                ImageContent(
                                    type="image",
                                    mimeType=mime_type,
                                    data=cast(str, data),
                                    metadata=json.dumps( # type: ignore
                                        {
                                            "cell_type": cell.cell_type,
                                            "cell_id": cell.id or global_cursor,
                                            "content_type": "output",
                                            "output_idx": output_idx
                                        }
                                    )
                                )
                            )
                        else:
                            # return other data types
                            if isinstance(data, list):
                                data = "\n".join(data)
                            result.append(
                                TextContent(
                                    type="text",
                                    text=json.dumps( # type: ignore
                                        {
                                            "cell_type": cell.cell_type,
                                            "cell_id": cell.id or global_cursor,
                                            "content_type": "output",
                                            "content": data,
                                            "output_idx": output_idx
                                        }
                                    )
                                )
                            )
                else:
                    result.append(
                        TextContent(
                            type="text",
                            text=json.dumps( # type: ignore
                                {
                                    "cell_type": cell.cell_type,
                                    "cell_id": cell.id or global_cursor,
                                    "content_type": "output",
                                    "output_idx": output_idx,
                                    "content": output.model_dump_json()
                                }
                            )
                        )
                    )

    # add metadata
    result.append(
        TextContent(
            type="text",
            text=json.dumps({
                "metadata": {
                    "total_cells": total_cells,
                    "current_range": {
                        "start": start,
                        "end": end
                    },
                    "cursor": new_cursor,
                    "has_more": has_more,
                    "reverse": reverse
                }
            })
        )
    )

    return result


@mcp.tool(
    name="list_jupyter_files",
    description="List files in the Jupyter workspace. To list files in the root directory, provide an empty string as file_path. To explore subdirectories, provide a relative path (e.g., 'subfolder/'). This tool helps you navigate and explore the workspace structure. Start with an empty path to discover the root directory contents first, then use specific paths to browse subdirectories."
)
def list_jupyter_files(
     file_path: Annotated[str, Field(default="", description="Relative path to list files from. Use empty string for root directory, or a path like 'subfolder/' for subdirectories")],
 ) -> list[TextContent]:
    # Validate path to prevent directory traversal
    if ".." in file_path:
        return [
            TextContent(
                type="text",
                text="Invalid path: Directory traversal are not allowed"
            )
        ]

    try:
        contents = cast(DirectoryContent, get_content(file_path).root)
    except Exception as e:
        return [
            TextContent(
                type="text",
                text=f"Failed to list files: {e}"
            )
        ]
    return [
        TextContent(
            type="text",
            text=content.model_dump_json()
        ) for content in contents.content
    ]



from mcp.server.fastmcp import FastMCP
import logging
import os
import argparse
import sys
import json
from tinydb import TinyDB, Query, where

# Set up logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
)
logger = logging.getLogger("tinydb-server")

# Global db instance
db = None
# This will hold the db file path received from argparse or set by tests
_db_file_path_for_get_db = None


def get_db(db_file_path_from_arg=None):
    """
    Gets the TinyDB instance. Initializes or re-initializes if the target db_file_path changes
    or if it's not initialized.
    The path for the DB is determined by:
    1. db_file_path_from_arg (if provided, typically from CLI or direct test call)
    2. _db_file_path_for_get_db (global, set by main() from CLI or by tests)
    """
    global db
    global _db_file_path_for_get_db

    target_db_path = db_file_path_from_arg if db_file_path_from_arg else _db_file_path_for_get_db

    if not target_db_path:
        # This case should ideally not be hit if main() correctly sets _db_file_path_for_get_db
        logger.error("Database file path not set. Using default 'db.json'.")
        target_db_path = 'db.json'

    # Update the global path if a new one was explicitly passed via argument (e.g. from test setup)
    if db_file_path_from_arg:
        _db_file_path_for_get_db = db_file_path_from_arg

    current_storage_path = None
    if db and db._storage:
        current_storage_path = db._storage.path

    if db is None or current_storage_path != target_db_path:
        if db:
            db.close()
        logger.info(f"Initializing TinyDB with file: {target_db_path}")
        db = TinyDB(target_db_path)
    return db

mcp = FastMCP()

def _build_query(query_params: dict) -> Query | None:
    """
    Builds a TinyDB Query object from a dictionary of parameters.
    Combines parameters with logical AND.
    Returns None if query_params is empty or None.
    """
    if not query_params:
        return None

    final_query: Query | None = None
    for key, value in query_params.items():
        current_condition = (Query()[key] == value)
        if final_query is None:
            final_query = current_condition
        else:
            final_query &= current_condition
    return final_query

@mcp.tool()
def create_table(table_name: str) -> str:
    """Creates a new table in the database.

    Args:
        table_name: The name of the table to create.

    Returns:
        A confirmation message.
    """
    current_db = get_db()
    current_db.table(table_name)
    return f"Table '{table_name}' created successfully."

@mcp.tool()
def insert_document(table_name: str, document: dict) -> str:
    """Inserts a document into the specified table.

    Args:
        table_name: The name of the table.
        document: The document to insert.

    Returns:
        The ID of the inserted document.
    """
    current_db = get_db()
    table = current_db.table(table_name)
    doc_id = table.insert(document)
    return str(doc_id)

@mcp.tool()
def query_documents(table_name: str, query_params: dict = None) -> str:
    """Queries documents from the specified table.

    Args:
        table_name: The name of the table.
        query_params: Optional. A dictionary of field-value pairs for exact matches.
                      Example: {"name": "John", "age": 30}

    Returns:
        A JSON string list of matching documents.
    """
    current_db = get_db()
    table = current_db.table(table_name)
    db_query = _build_query(query_params)

    if db_query is not None:
        results = table.search(db_query)
    else:
        results = table.all()
    return json.dumps(results)

@mcp.tool()
def update_documents(table_name: str, query_params: dict, update_data: dict) -> str:
    """Updates documents in the specified table.

    Args:
        table_name: The name of the table.
        query_params: A dictionary of field-value pairs to identify documents to update.
        update_data: A dictionary of field-value pairs to update.

    Returns:
        The number of updated documents.
    """
    current_db = get_db()
    table = current_db.table(table_name)
    db_query = _build_query(query_params)

    if db_query is None:
        return "Error: Query parameters are required for update operations."

    updated_ids = table.update(update_data, db_query)
    return str(len(updated_ids))


@mcp.tool()
def delete_documents(table_name: str, query_params: dict) -> str:
    """Deletes documents from the specified table.

    Args:
        table_name: The name of the table.
        query_params: A dictionary of field-value pairs to identify documents to delete.

    Returns:
        The number of deleted documents.
    """
    current_db = get_db()
    table = current_db.table(table_name)
    db_query = _build_query(query_params)

    if db_query is None:
        return "Error: Query parameters are required for delete operations."

    deleted_ids = table.remove(db_query)
    return str(len(deleted_ids))

@mcp.tool()
def purge_table(table_name: str) -> str:
    """Removes all documents from the specified table.

    Args:
        table_name: The name of the table to purge.

    Returns:
        A confirmation message.
    """
    current_db = get_db()
    table = current_db.table(table_name)
    table.truncate()
    return f"Table '{table_name}' purged successfully."

@mcp.tool()
def drop_table(table_name: str) -> str:
    """Drops the specified table from the database.

    Args:
        table_name: The name of the table to drop.

    Returns:
        A confirmation message.
    """
    current_db = get_db()
    current_db.drop_table(table_name)
    return f"Table '{table_name}' dropped successfully."

def server_main():
    parser = argparse.ArgumentParser(description="MCP TinyDB Server")
    parser.add_argument(
        "--db-file",
        type=str,
        default="db.json",
        help="Path to the TinyDB JSON file.",
    )
    args = parser.parse_args(sys.argv[1:])  # Use only arguments after script name

    global _db_file_path_for_get_db
    _db_file_path_for_get_db = args.db_file

    logger.info(f"Starting tinydb-server. Database file: {args.db_file}")

    get_db(args.db_file)

    mcp.run('stdio')

if __name__ == "__main__":
    server_main()


from setuptools import setup, find_packages

setup(
    name="mathematica-check",
    version="0.1.0",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "mcp-mathematica-check=mathematica_check.server:main",
        ],
    },
    install_requires=[
        "mcp[cli]>=1.0.0",
    ],
)



import os
import logging
from typing import List
from fastapi import HTTPException


def validate_command(
    command: str,
    allowed_commands: List[str],
    allowed_paths: List[str],
    commands_blacklist: List[str],
    arguments_blacklist: List[str]
) -> None:
    """Validate if a command is allowed to be executed

    Args:
        command: Command to validate
        allowed_commands: List of allowed commands
        allowed_paths: List of allowed paths
        commands_blacklist: List of blacklisted commands
        arguments_blacklist: List of blacklisted arguments

    Raises:
        HTTPException: If the command validation fails
    """
    # Strip the command of any leading/trailing whitespace
    command = command.strip()

    # Check if the command is empty
    if not command:
        error_msg = "Command cannot be empty"
        logging.error("Command validation failed: %s", error_msg)
        raise HTTPException(
            status_code=400, detail=f"Command validation failed: {error_msg}")

    # Split the command into parts (command and arguments)
    parts = command.split()
    base_command = parts[0]

    # Check if the command is in the blacklist
    for blacklisted_cmd in commands_blacklist:
        if base_command == blacklisted_cmd or base_command.endswith(f"/{blacklisted_cmd}"):
            error_msg = f"Command '{base_command}' is blacklisted"
            logging.error("Command validation failed: %s", error_msg)
            raise HTTPException(
                status_code=400, detail=f"Command validation failed: {error_msg}")

    # Check for blacklisted arguments
    for arg in parts[1:]:
        for blacklisted_arg in arguments_blacklist:
            if arg == blacklisted_arg:
                error_msg = f"Argument '{arg}' is blacklisted"
                logging.error("Command validation failed: %s", error_msg)
                raise HTTPException(
                    status_code=400, detail=f"Command validation failed: {error_msg}")

    # If allowed_commands is provided, check if the command is in the list
    if allowed_commands:
        is_allowed = False
        for allowed_cmd in allowed_commands:
            if command.startswith(allowed_cmd):
                is_allowed = True
                break

        if not is_allowed:
            error_msg = f"Command '{command}' is not in the allowed commands list"
            logging.error("Command validation failed: %s", error_msg)
            raise HTTPException(
                status_code=400, detail=f"Command validation failed: {error_msg}")

    # If allowed_paths is provided, check if the command operates on allowed paths
    if allowed_paths and any(path_arg for path_arg in parts[1:] if not path_arg.startswith("-")):
        # Extract potential path arguments (those not starting with -)
        potential_paths = [arg for arg in parts[1:] if not arg.startswith("-")]

        for path_arg in potential_paths:
            path_allowed = False

            # Skip empty path arguments
            if not path_arg.strip():
                continue

            # Check if the path is within any of the allowed paths
            for allowed_path in allowed_paths:
                # Normalize paths for comparison
                norm_path_arg = os.path.normpath(path_arg)
                norm_allowed_path = os.path.normpath(allowed_path)

                if (norm_path_arg.startswith(norm_allowed_path) or
                        os.path.abspath(norm_path_arg).startswith(norm_allowed_path)):
                    path_allowed = True
                    break

            if not path_allowed:
                error_msg = f"Path '{path_arg}' is not in the allowed paths list"
                logging.error("Command validation failed: %s", error_msg)
                raise HTTPException(
                    status_code=400, detail=f"Command validation failed: {error_msg}")

    # If we get here, the command is valid
    return



"""NEMAD MCP Server"""

import argparse
from . import server


def main():
    """Main CLI entry point for the NEMAD MCP server"""
    parser = argparse.ArgumentParser(description="NEMAD MCP Server")
    parser.add_argument(
        "--transport",
        choices=["stdio", "sse"],
        default="stdio",
        help="Transport method (default: stdio)",
    )
    
    args = parser.parse_args()
    
    # Run the server
    server.main(args.transport)


__all__ = ["main"]



import os
import logging
import requests
from typing import Any, Dict, List, Optional
from mcp.server.fastmcp import FastMCP
from mcp.types import TextContent

logger = logging.getLogger(__name__)

# Create the MCP server
mcp = FastMCP()

# Global cache for search results
_search_cache: Dict[str, List[Dict[str, Any]]] = {}

NEMAD_BASE_URL = "https://nemad.org/api"


def get_api_key() -> str:
    """Get API key from environment variable"""
    api_key = os.getenv("NEMAD_API_KEY")
    if not api_key:
        raise ValueError("NEMAD_API_KEY environment variable is required")
    return api_key


def make_request(url: str, params: Dict[str, Any]) -> Dict[str, Any]:
    """Make authenticated request to NEMAD API"""
    headers = {"X-API-Key": get_api_key(), "accept": "application/json"}

    response = requests.get(url, headers=headers, params=params)
    response.raise_for_status()
    return response.json()


@mcp.tool()
def nemad_search(
    database_type: str, elements: str, exact_match: bool = False, limit: int = -1
) -> TextContent:
    """Search for materials by elements in NEMAD database.

    Args:
        database_type: Type of database (magnetic, thermoelectric, or superconductor)
        elements: Comma-separated list of elements (e.g., "Fe,Co,Ni")
        exact_match: If true, return materials with exactly these elements only
        limit: Number of results to return (-1 for all results)

    Returns:
        Search results with count and material information
    """
    try:
        # Validate database type
        valid_db_types = ["magnetic", "thermoelectric", "superconductor"]
        if database_type not in valid_db_types:
            return TextContent(
                type="text",
                text=f"Error: database_type must be one of {valid_db_types}",
            )

        # Construct URL and parameters
        url = f"{NEMAD_BASE_URL}/{database_type}/search"
        params = {"elements": elements, "exact_match": exact_match, "limit": limit}

        # Make request
        data = make_request(url, params)

        # Cache the results for later reading
        cache_key = f"{database_type}_{elements}_{exact_match}_{limit}"
        _search_cache[cache_key] = data.get("results", [])

        # Format response
        count = data.get("count", 0)
        results = data.get("results", [])

        response_text = "NEMAD Search Results\n"
        response_text += f"Database: {database_type}\n"
        response_text += f"Elements: {elements}\n"
        response_text += f"Exact match: {exact_match}\n"
        response_text += f"Total materials found: {count}\n"
        response_text += f"Results returned: {len(results)}\n"
        response_text += f"Cache key: {cache_key}\n\n"
        response_text += f"Use nemad_read_results with cache_key='{cache_key}' to read specific ranges of results."

        return TextContent(type="text", text=response_text)

    except Exception as e:
        logger.error(f"Error in nemad_search: {e}")
        return TextContent(type="text", text=f"Error: {str(e)}")


@mcp.tool()
def nemad_formula_search(
    database_type: str, formula: str, limit: int = -1
) -> TextContent:
    """Search for materials by exact chemical formula in NEMAD database.

    Args:
        database_type: Type of database (magnetic, thermoelectric, or superconductor)
        formula: Exact chemical formula to search for (e.g., "Fe2O3")
        limit: Number of results to return (-1 for all results)

    Returns:
        Search results with count and material information
    """
    try:
        # Validate database type
        valid_db_types = ["magnetic", "thermoelectric", "superconductor"]
        if database_type not in valid_db_types:
            return TextContent(
                type="text",
                text=f"Error: database_type must be one of {valid_db_types}",
            )

        # Construct URL and parameters
        url = f"{NEMAD_BASE_URL}/{database_type}/formula"
        params = {"formula": formula, "limit": limit}

        # Make request
        data = make_request(url, params)

        # Cache the results for later reading
        cache_key = f"{database_type}_formula_{formula}_{limit}"
        _search_cache[cache_key] = data.get("results", [])

        # Format response
        count = data.get("count", 0)
        results = data.get("results", [])

        response_text = "NEMAD Formula Search Results\n"
        response_text += f"Database: {database_type}\n"
        response_text += f"Formula: {formula}\n"
        response_text += f"Total materials found: {count}\n"
        response_text += f"Results returned: {len(results)}\n"
        response_text += f"Cache key: {cache_key}\n\n"
        response_text += f"Use nemad_read_results with cache_key='{cache_key}' to read specific ranges of results."

        return TextContent(type="text", text=response_text)

    except Exception as e:
        logger.error(f"Error in nemad_formula_search: {e}")
        return TextContent(type="text", text=f"Error: {str(e)}")


@mcp.tool()
def nemad_read_results(
    cache_key: str, start_index: int = 0, end_index: Optional[int] = None
) -> TextContent:
    """Read specific range of cached search results.

    Args:
        cache_key: The cache key from previous search results
        start_index: Starting index (0-based, inclusive)
        end_index: Ending index (0-based, exclusive). If None, read to end.

    Returns:
        Detailed information for materials in the specified range
    """
    try:
        # Check if cache key exists
        if cache_key not in _search_cache:
            available_keys = list(_search_cache.keys())
            return TextContent(
                type="text",
                text=f"Error: Cache key '{cache_key}' not found.\nAvailable keys: {available_keys}",
            )

        results = _search_cache[cache_key]
        total_count = len(results)

        # Validate indices
        if start_index < 0:
            start_index = 0
        if start_index >= total_count:
            return TextContent(
                type="text",
                text=f"Error: start_index {start_index} is out of range. Total results: {total_count}",
            )

        if end_index is None:
            end_index = total_count
        else:
            end_index = min(end_index, total_count)

        if end_index <= start_index:
            return TextContent(
                type="text",
                text=f"Error: end_index {end_index} must be greater than start_index {start_index}",
            )

        # Extract the requested range
        selected_results = results[start_index:end_index]

        # Format response
        response_text = f"NEMAD Results (Range {start_index} to {end_index - 1})\n"
        response_text += f"Cache key: {cache_key}\n"
        response_text += f"Total available results: {total_count}\n"
        response_text += f"Showing results: {len(selected_results)}\n\n"

        for i, material in enumerate(selected_results):
            actual_index = start_index + i
            response_text += f"Result #{actual_index + 1}:\n"
            response_text += (
                f"Material Name: {material.get('Material_Name', 'Unknown')}\n"
            )

            for key, value in material.items():
                if key != "Material_Name":
                    response_text += f"{key}: {value}\n"
            response_text += "\n" + "-" * 50 + "\n\n"

        return TextContent(type="text", text=response_text)

    except Exception as e:
        logger.error(f"Error in nemad_read_results: {e}")
        return TextContent(type="text", text=f"Error: {str(e)}")


@mcp.tool()
def nemad_list_cache() -> TextContent:
    """List all available cached search results.

    Returns:
        List of all cache keys with basic information
    """
    if not _search_cache:
        return TextContent(
            type="text",
            text="No cached search results available. Perform a search first.",
        )

    response_text = "Available Cached Search Results:\n\n"
    for cache_key, results in _search_cache.items():
        response_text += f"Cache Key: {cache_key}\n"
        response_text += f"Results Count: {len(results)}\n"
        response_text += "-" * 30 + "\n"

    return TextContent(type="text", text=response_text)


def main(transport: str = "stdio"):
    """Main entry point for the NEMAD MCP server"""
    logger.info("Starting NEMAD MCP server")
    mcp.run(transport)



import plotly.graph_objects as go
import numpy as np
# from ase.dft.band_structure import BandStructure
from ase.spectrum.band_structure import BandStructure
from gpaw.spinorbit import soc_eigenstates
from ase.dft.dos import DOS
import matplotlib.colors as mcolors


def plot_band(
        calc,
        emin=-6,
        emax=3,
        ylabel='Energy (eV)',
        colors=[
            'blue',
            'red'],
    labels=[
            'w/o SOC',
            'w/ SOC'],
        soc=False,
        only_show_soc=False):
    """
    Plot the band structure with and without SOC, but non spin-polarization
    The band structure with SOC is calculated in a perturbative way.

    Parameters:
    calc (Calculator): The calculator object.
    emin (float): The minimum y-axis value. Default is -6.
    emax (float): The maximum y-axis value. Default is 3.
    ylabel (str): The label for the y-axis. Default is 'Energy (eV)'.
    colors (list[str]): The colors of the band structure. Default is ['blue', 'red'].
    labels (list[str]): The labels of the band structure. Default is ['w/o SOC', 'w/ SOC'].
    soc (bool): Whether to add the band structure with SOC. Default is False.
    only_show_soc (bool): Whether to only show the band structure with SOC. Default is False.
    """

    # the ordinary band structure plot
    bands = calc.band_structure()
    bands = bands.subtract_reference()  # shift the band to Fermi level
    (x, X, xlabels) = bands.path.get_linear_kpoint_axis()
    fig = go.Figure()

    # plot the band structure without SOC
    if not only_show_soc or not soc:
        for i in range(bands.energies.shape[2]):
            fig.add_trace(go.Scatter(
                x=x,
                y=bands.energies[0, :, i],
                mode='lines',
                line=dict(color=colors[0]),
                name=labels[0] if i == 0 else '',
                showlegend=(i == 0)
            ))

    # plot the band structure with SOC
    if soc:
        # calculate the band with SOC in a perturbative way
        soc_calc = soc_eigenstates(calc)
        energies_soc = soc_calc.eigenvalues()
        energies_soc -= soc_calc.fermi_level

        # construct the bandstructure object with SOC
        energies_soc = np.array(energies_soc)
        nkpt = energies_soc.shape[0]
        nband = energies_soc.shape[1]
        energies_soc = energies_soc.reshape(1, nkpt, nband)
        bands_soc = BandStructure(bands.path, energies_soc, reference=0)

        # Plot the SOC band structure
        for i in range(bands_soc.energies.shape[2]):
            fig.add_trace(go.Scatter(
                x=x,
                y=bands_soc.energies[0, :, i],
                mode='lines',
                line=dict(color=colors[1], dash='dash'),
                name=labels[1] if i == 0 else '',
                showlegend=(i == 0)
            ))

    # Update layout
    fig.update_layout(
        xaxis_title='k-points',
        yaxis_title=ylabel,
        yaxis_range=[emin, emax],
        xaxis_tickvals=X,
        xaxis_ticktext=xlabels,
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1),
    )

    # Add vertical lines for special k-points
    for x in bands.path.special_points.values():
        fig.add_vline(x=x, line_width=1, line_dash="dash", line_color="gray")

    # save the band structure to json file
    bands.write('bandstructure.json')
    if soc:
        bands_soc.write('bandstructure_soc.json')

    return fig


def plot_band_projection(calc, locfun, scale_factor=100, emin=-6, emax=3,
                         ylabel='Energy (eV)', xlabel='K-path', colors=['blue', 'red'],
                         labels=['Bands', 'Projected Bands']):
    """
    Plot the band structure with projections
    non-spin polarized, no SOC
    """
    bands = calc.band_structure()
    bands = bands.subtract_reference()  # shift the band to Fermi level
    energies = bands.energies
    path = bands.path
    (x, X, xlabels) = path.get_linear_kpoint_axis()

    fig = go.Figure()

    if locfun:
        proj_kni = calc.get_projections(locfun, spin=0)  # Assuming spin-up, change to 1 for spin-down if needed
        projections = np.abs(proj_kni)**2
        projections_sum = np.sum(projections, axis=2)

        # Plot bands without projections as background
        for band in range(energies.shape[2]):
            fig.add_trace(go.Scatter(
                x=x, y=energies[0, :, band],
                mode='lines', line=dict(color=colors[0]),
                name=labels[0] if band == 0 else '',
                showlegend=(band == 0)
            ))

        # Plot the projected bands
        for band in range(energies.shape[2]):
            sizes = projections_sum[:, band] * scale_factor
            fig.add_trace(go.Scatter(
                x=x, y=energies[0, :, band],
                mode='markers',
                marker=dict(size=sizes, color=colors[1], opacity=0.5),
                name=labels[1] if band == 0 else '',
                showlegend=(band == 0)
            ))

        # Add vertical lines for high-symmetry points
        for label_position in X:
            fig.add_vline(x=label_position, line_dash="dash", line_color="gray", line_width=1)

    else:
        raise ValueError('locfun is not provided')
        # # If no projections, just plot regular band structure
        # for band in range(energies.shape[2]):
        #     fig.add_trace(go.Scatter(
        #         x=x, y=energies[0, :, band],
        #         mode='lines', line=dict(color=colors[0]),
        #         name=labels[0] if band == 0 else '',
        #         showlegend=(band == 0)
        #     ))

    # Update layout
    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        yaxis_range=[emin, emax],
        xaxis_tickvals=X,
        xaxis_ticktext=xlabels,
        title='Band Structure',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add a horizontal line at Fermi level
    fig.add_hline(y=0, line_dash="dash", line_color="gray")

    return fig


def plot_band_spinpol(calc, emin=-3, emax=3, soc=True, only_show_soc=True, show_spinprojection=True,
                      ylabel='Energy (eV)', xlabel='K-path', marker_size=5, marker_style='circle'):
    """
    Plot the band structure with spin polarization,
    with and without SOC
    """
    bands = calc.band_structure()
    efermi = bands.reference
    bands = bands.subtract_reference()  # shift the band to Fermi level

    (x, X, xlabels) = bands.path.get_linear_kpoint_axis()

    fig = go.Figure()

    if not soc:
        only_show_soc = False
        show_spinprojection = False

    if not only_show_soc or not soc:
        # Plot the band structure with spin polarization
        for spin in range(2):  # Assuming spin-polarized calculation
            for band in range(bands.energies.shape[2]):
                fig.add_trace(go.Scatter(
                    x=x, y=bands.energies[spin, :, band],
                    mode='lines',
                    line=dict(color='blue' if spin == 0 else 'red'),
                    name=f'Spin {"up" if spin == 0 else "down"}' if band == 0 else '',
                    showlegend=(band == 0)
                ))

    if soc:
        # Calculate the band with SOC in a perturbative way
        soc_calc = soc_eigenstates(calc)
        energies_soc = soc_calc.eigenvalues().T
        spin_projection = soc_calc.spin_projections()
        energies_soc -= efermi
        s_nk = (spin_projection[:, :, 2].T + 1.0) / 2.0

        if show_spinprojection:
            # Plot the spin projected band structure
            # Plot each band segment separately
            for i in range(len(energies_soc)):
                # Split data at high symmetry points
                split_points = [0] + [np.where(x == pos)[0][0] for pos in X[1:-1]] + [len(x) - 1]

                for j in range(len(split_points) - 1):
                    start_idx = split_points[j]
                    end_idx = split_points[j + 1] + 1

                    segment_x = x[start_idx:end_idx]
                    segment_y = energies_soc[i, start_idx:end_idx]
                    segment_colors = s_nk[i, start_idx:end_idx]

                    fig.add_trace(go.Scatter(
                        x=segment_x,
                        y=segment_y,
                        mode='lines+markers',
                        marker=dict(
                            size=marker_size,
                            color=segment_colors,
                            colorscale='RdBu_r',
                            colorbar=dict(
                                title=dict(
                                    text='Spin projection',
                                    side='right'
                                ),
                                tickmode='array',
                                tickvals=[0, 0.5, 1],
                                ticktext=['Down', 'Mixed', 'Up'],
                                x=1.2,
                                thickness=15,
                                ticklabelposition='outside right'
                            ) if (i == 0 and j == 0) else None,
                            symbol=marker_style,
                            opacity=0.5
                        ),
                        line=dict(color='gray'),
                        name='SOC bands' if (i == 0 and j == 0) else None,
                        showlegend=(i == 0 and j == 0)
                    ))
        else:
            # only plot the SOC band
            for band in range(energies_soc.shape[0]):
                fig.add_trace(go.Scatter(
                    x=x,
                    y=energies_soc[band, :].T,
                    mode='lines',
                    line=dict(color='gray'),
                    name='SOC bands',
                    showlegend=(band == 0)
                ))

    # Update layout
    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        yaxis_range=[emin, emax],
        xaxis_tickvals=X,
        xaxis_ticktext=xlabels,
        title='Band Structure with Spin Polarization',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add vertical lines for high-symmetry points
    for label_position in X:
        fig.add_vline(x=label_position, line_dash="dash", line_color="gray", line_width=1)

    # Add a horizontal line at Fermi level
    fig.add_hline(y=0, line_dash="dash", line_color="gray")

    return fig


def plot_band_spinpol_projection(calc, locfun, emin=-6, emax=3, colors=['blue', 'red'],
                                 labels=['Spin up', 'Spin down'],
                                 scale_factor=10, xlabel='K-path', ylabel='Energy (eV)',
                                 only_show_projection=False):
    """
    Plot the band structure with projections
    Spin polarized, no SOC
    """
    # sigma =  0.1
    # locfun = [[0, 1, sigma], [0, 0, sigma], [0, 2, sigma], [0, 3, sigma]]
    # colors = ['blue', 'red']
    # labels = ['Spin up', 'Spin down']
    # emin = -6
    # emax = 3
    # ylabel = 'Energy (eV)'
    # xlabel = 'K-path'
    # scale_factor = 10

    bands = calc.band_structure()
    bands = bands.subtract_reference()  # shift the band to Fermi level
    energies = bands.energies
    path = bands.path
    (x, X, xlabels) = path.get_linear_kpoint_axis()

    fig = go.Figure()

    if locfun:
        proj_kni_up = calc.get_projections(locfun, spin=0)  # Assuming spin-up, change to 1 for spin-down if needed
        proj_kni_down = calc.get_projections(locfun, spin=1)
        # sum over all atom projections
        # note: it is possible to split px py pz here
        projections_up = np.sum(np.abs(proj_kni_up)**2, axis=2)
        projections_down = np.sum(np.abs(proj_kni_down)**2, axis=2)

        # Plot the projected bands
        for band in range(energies.shape[2]):
            sizes_up = projections_up[:, band] * scale_factor
            sizes_down = projections_down[:, band] * scale_factor

            # Split data at high symmetry points
            split_points = [0] + [np.where(x == pos)[0][0] for pos in X[1:-1]] + [len(x) - 1]

            for j in range(len(split_points) - 1):
                start_idx = split_points[j]
                end_idx = split_points[j + 1] + 1

                segment_x = x[start_idx:end_idx]
                segment_y_up = energies[0, start_idx:end_idx, band]
                segment_y_down = energies[1, start_idx:end_idx, band]
                segment_sizes_up = sizes_up[start_idx:end_idx]
                segment_sizes_down = sizes_down[start_idx:end_idx]

                fig.add_trace(go.Scatter(
                    x=segment_x,
                    y=segment_y_up,
                    mode='lines+markers',
                    marker=dict(size=segment_sizes_up, color=colors[0], opacity=0.5),
                    line=dict(color=colors[0]),
                    name=labels[0] if (band == 0 and j == 0) else None,
                    showlegend=(band == 0 and j == 0)
                ))
                fig.add_trace(go.Scatter(
                    x=segment_x,
                    y=segment_y_down,
                    mode='lines+markers',
                    marker=dict(size=segment_sizes_down, color=colors[1], opacity=0.5),
                    line=dict(color=colors[1]),
                    name=labels[1] if (band == 0 and j == 0) else None,
                    showlegend=(band == 0 and j == 0)
                ))

        # Add vertical lines for high-symmetry points
        for label_position in X:
            fig.add_vline(x=label_position, line_dash="dash", line_color="gray", line_width=1)
    else:
        raise ValueError('locfun is not provided')

    if only_show_projection == False or locfun is None:
        # If no projections, just pslot regular band structure
        for spin in range(energies.shape[0]):
            for band in range(energies.shape[2]):
                fig.add_trace(go.Scatter(
                    x=x, y=energies[spin, :, band],
                    mode='lines', line=dict(color=colors[spin]),
                    name=labels[spin] if band == 0 else '',
                    showlegend=(band == 0)
                ))

    # Update layout
    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        yaxis_range=[emin, emax],
        xaxis_tickvals=X,
        xaxis_ticktext=xlabels,
        title='Band Structure',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add a horizontal line at Fermi level
    fig.add_hline(y=0, line_dash="dash", line_color="gray")

    return fig


def plot_dos(calc, emin=-6, emax=3, xlabel='Energy (eV)', ylabel='DOS (states/eV)',
             npts=1000, width=0.1, color='black', label='Total DOS'):
    """
    Extract the total DOS from the GPAW calculator and plot it
    the calculator can be a ground state or a band calculation
    non-spin polarized, no SOC

    Parameters:
    calc (Calculator): The calculator object.
    emin (float): The minimum energy value. Default is -6.
    emax (float): The maximum energy value. Default is 3.
    xlabel (str): The x-axis label. Default is 'Energy (eV)'.
    ylabel (str): The y-axis label. Default is 'DOS (states/eV)'.
    npts (int): Number of points for DOS calculation. Default is 1000.
    width (float): Width parameter for DOS calculation. Default is 0.1.
    color (str): Color for the DOS plot. Default is 'black'.
    label (str): Label for the DOS plot. Default is 'Total DOS'.
    """

    # Calculate total DOS
    dos = DOS(calc, npts=npts, width=width)
    energies = dos.energies
    dos_values = dos.get_dos()

    # Create the plot
    fig = go.Figure()

    fig.add_trace(go.Scatter(
        x=energies,
        y=dos_values,
        mode='lines',
        line=dict(color=color),
        name=label
    ))

    # Set axis limits
    mask = (energies >= emin) & (energies <= emax)
    ymax = max(dos_values[mask]) * 1.1  # Add 10% margin for better visibility

    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        xaxis_range=[emin, emax],
        yaxis_range=[0, ymax],
        title='Density of States',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add a vertical line at Fermi level (E=0)
    fig.add_vline(x=0, line_dash="dash", line_color="gray")

    return fig


def plot_dos_projection(calc, locfun, emin=-6, emax=3, xlabel='Energy (eV)',
                        ylabel='DOS (states/eV)', npts=1000, color='black'):
    """
    Extract the total DOS and PDOS from the GPAW calculator and plot them
    the calculator can be a ground state or a band calculation
    non-spin polarized, no SOC

    Parameters:
    calc (Calculator): The calculator object.
    locfun (list): List of [atom_index, angular, sigma] for PDOS calculation.
    emin (float): The minimum energy value. Default is -6.
    emax (float): The maximum energy value. Default is 3.
    xlabel (str): The x-axis label. Default is 'Energy (eV)'.
    ylabel (str): The y-axis label. Default is 'DOS (states/eV)'.
    npts (int): Number of points for DOS calculation. Default is 1000.
    color (str): Base color for the plots. Default is 'black'.
    """

    if not locfun:
        raise ValueError('locfun is not provided')
        # return plot_dos(calc, emin=emin, emax=emax, xlabel=xlabel, ylabel=ylabel,
        #                 npts=npts, width=0.1, color=color, label='Total DOS')

    width = locfun[0][2]  # read the width from locfun

    # Calculate total DOS
    dos = DOS(calc, npts=npts, width=width)
    energies = dos.energies
    dos_values = dos.get_dos()

    # Calculate PDOS
    pdos_values = []
    for loc in locfun:
        a, l, _ = loc
        angular = 's' if l == 0 else 'p' if l == 1 else 'd' if l == 2 else 'f' if l == 3 else None
        _, pdos = calc.get_orbital_ldos(a=a, angular=angular, npts=npts, width=width)
        pdos_values.append(pdos)

    # Sum up the PDOS
    pdos_sum = sum(pdos_values)

    # Create the plot
    fig = go.Figure()

    fig.add_trace(go.Scatter(
        x=energies,
        y=dos_values,
        mode='lines',
        line=dict(color=color),
        name='Total DOS'
    ))

    fig.add_trace(go.Scatter(
        x=energies,
        y=pdos_sum,
        mode='lines',
        line=dict(color=color, dash='dash'),
        name='PDOS'
    ))

    # Set axis limits
    mask = (energies >= emin) & (energies <= emax)
    ymax = max(dos_values[mask]) * 1.1  # Add 10% margin for better visibility

    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        xaxis_range=[emin, emax],
        yaxis_range=[0, ymax],
        title='Density of States with Projections',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add a vertical line at Fermi level (E=0)
    fig.add_vline(x=0, line_dash="dash", line_color="gray")

    return fig


def plot_dos_spinpol(calc, emin=-6, emax=3, npts=1000, width=0.1, xlabel='Energy (eV)',
                     ylabel='DOS (states/eV)', colors=['red', 'blue'],
                     labels=['Spin up', 'Spin down']):
    """
    Extract the total DOS from the GPAW calculator and plot it
    for spin-polarized calculations

    Parameters:
    calc (Calculator): The calculator object.
    emin (float): The minimum energy value. Default is -6.
    emax (float): The maximum energy value. Default is 3.
    npts (int): Number of points for DOS calculation. Default is 1000.
    width (float): Width parameter for DOS calculation. Default is 0.1.
    xlabel (str): The x-axis label. Default is 'Energy (eV)'.
    ylabel (str): The y-axis label. Default is 'DOS (states/eV)'.
    colors (list): Colors for spin up and down. Default is ['red', 'blue'].
    labels (list): Labels for spin up and down. Default is ['Spin up', 'Spin down'].
    """
    if calc.get_number_of_spins() == 1:
        raise ValueError("The calculation is not spin polarized")
    elif calc.get_number_of_spins() == 2:
        pass
    else:
        raise ValueError(f"Unsupported number of spins: {calc.get_number_of_spins()}. Use 1 or 2.")

    # Convert named colors to hex if necessary
    def to_hex(color):
        if color.startswith('#'):
            return color
        return mcolors.to_hex(color)

    colors = [to_hex(color) for color in colors]

    # Calculate total DOS
    dos = DOS(calc, npts=npts, width=width)
    energies = dos.energies
    dos_up = dos.get_dos(spin=0)
    dos_down = dos.get_dos(spin=1)

    # Create the plot
    fig = go.Figure()

    # Spin up DOS
    fig.add_trace(go.Scatter(
        x=energies,
        y=dos_up,
        fill='tozeroy',
        fillcolor=f'rgba({int(colors[0][1:3], 16)},{int(colors[0][3:5], 16)},{int(colors[0][5:7], 16)},0.5)',
        line=dict(color=colors[0], width=1.5),
        name=labels[0]
    ))

    # Spin down DOS (negated for visual separation)
    fig.add_trace(go.Scatter(
        x=energies,
        y=-dos_down,
        fill='tozeroy',
        fillcolor=f'rgba({int(colors[1][1:3], 16)},{int(colors[1][3:5], 16)},{int(colors[1][5:7], 16)},0.5)',
        line=dict(color=colors[1], width=1.5),
        name=labels[1]
    ))

    # Set axis limits
    mask = (energies >= emin) & (energies <= emax)
    ymax = max(max(dos_up[mask]), max(dos_down[mask])) * 1.1  # Add 10% margin for better visibility

    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        xaxis_range=[emin, emax],
        yaxis_range=[-ymax, ymax],
        title='Spin-Polarized Density of States',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add a vertical line at Fermi level (E=0)
    fig.add_vline(x=0, line_dash="dash", line_color="gray")

    return fig


def plot_dos_spinpol_projection(calc, locfun, emin=-6, emax=3, npts=1000, xlabel='Energy (eV)',
                                ylabel='DOS (states/eV)', colors=['red', 'blue'],
                                labels=['Spin up', 'Spin down']):
    if not locfun:
        raise ValueError('locfun is not provided')

    width = locfun[0][2]  # read the width from locfun

    if calc.get_number_of_spins() == 1:
        raise ValueError("The calculation is not spin polarized")
    elif calc.get_number_of_spins() == 2:
        pass
    else:
        raise ValueError(f"Unsupported number of spins: {calc.get_number_of_spins()}. Use 1 or 2.")

    # Convert named colors to hex if necessary
    def to_hex(color):
        if color.startswith('#'):
            return color
        return mcolors.to_hex(color)

    colors = [to_hex(color) for color in colors]

    # Calculate total DOS
    dos = DOS(calc, npts=npts, width=width)
    energies = dos.energies
    dos_up = dos.get_dos(spin=0)
    dos_down = dos.get_dos(spin=1)

    # Create the plot
    fig = go.Figure()

    # Spin up DOS
    fig.add_trace(go.Scatter(
        x=energies,
        y=dos_up,
        fill='tozeroy',
        fillcolor=f'rgba({int(colors[0][1:3], 16)},{int(colors[0][3:5], 16)},{int(colors[0][5:7], 16)},0.5)',
        line=dict(color=colors[0], width=1.5),
        name=labels[0]
    ))

    # Spin down DOS (negated for visual separation)
    fig.add_trace(go.Scatter(
        x=energies,
        y=-dos_down,
        fill='tozeroy',
        fillcolor=f'rgba({int(colors[1][1:3], 16)},{int(colors[1][3:5], 16)},{int(colors[1][5:7], 16)},0.5)',
        line=dict(color=colors[1], width=1.5),
        name=labels[1]
    ))

    # Calculate PDOS
    pdos_values_up = []
    pdos_values_down = []
    for loc in locfun:
        a, l, _ = loc
        angular = 's' if l == 0 else 'p' if l == 1 else 'd' if l == 2 else 'f' if l == 3 else None
        _, pdos_up = calc.get_orbital_ldos(a=a, angular=angular, npts=npts, width=width, spin=0)
        _, pdos_down = calc.get_orbital_ldos(a=a, angular=angular, npts=npts, width=width, spin=1)
        pdos_values_up.append(pdos_up)
        pdos_values_down.append(pdos_down)

    # Sum up the PDOS
    pdos_sum_up = sum(pdos_values_up)
    pdos_sum_down = sum(pdos_values_down)

    fig.add_trace(go.Scatter(
        x=energies,
        y=pdos_sum_up,
        mode='lines',
        line=dict(color=colors[0], dash='dash'),
        name=labels[0]
    ))

    fig.add_trace(go.Scatter(
        x=energies,
        y=-pdos_sum_down,
        mode='lines',
        line=dict(color=colors[1], dash='dash'),
        name=labels[1]
    ))

    # Set axis limits
    mask = (energies >= emin) & (energies <= emax)
    ymax = max(dos_up[mask]) * 1.1  # Add 10% margin for better visibility
    ymin = min(-dos_down[mask]) * 1.1

    fig.update_layout(
        xaxis_title=xlabel,
        yaxis_title=ylabel,
        xaxis_range=[emin, emax],
        yaxis_range=[ymin, ymax],
        title='Density of States with Projections',
        legend=dict(orientation='h', yanchor='bottom', y=1.02, xanchor='right', x=1)
    )

    # Add a vertical line at Fermi level (E=0)
    fig.add_vline(x=0, line_dash="dash", line_color="gray")

    # fig.show()
    return fig



from ase.atoms import Atoms
from gpaw import GPAW, restart
from ase.optimize import BFGS
import time
import uuid
import os
import sys
import psutil
from .data_class import CalculationInfo, EnhancedAtoms, CalculationProject
import plotly.io as pio


GPAW_TEMP_FOLDER_PATH = "working/temp"

FORCE_THRESHOLD = 0.001

_MAX_TRY = 10

MAX_COMPUTATION_TIME = 3600


def resume_relax_calculation(ckpt_file_path: str,
                             trajectory_file_path: str,
                             steps=100,
                             force_threshold=FORCE_THRESHOLD):
    """Resume relax calculation from a checkpoint file."""
    if os.path.exists(ckpt_file_path):
        # Restart from checkpoint
        atoms, calc = restart(ckpt_file_path)
        atoms = EnhancedAtoms(atoms)
        is_converged = atoms.check_convergence(force_threshold=force_threshold)
        if is_converged:
            # The checkpoint is a converged calculation. Output results
            atoms.calc = calc
            return atoms
        else:
            print(
                "The previous calculation is not converged. Resume the calculation now.")
            optimizer = BFGS(atoms, trajectory=trajectory_file_path)
            # Enable checkpointing in the optimizer
            optimizer.attach(lambda: calc.write(ckpt_file_path), interval=1)
            optimizer.run(fmax=force_threshold, steps=steps)
            return atoms


def relax_calculation(atoms: Atoms,
                      is_bulk: bool,
                      kpoints: list[int],
                      spinpol: bool,
                      status_file_path: str,
                      log_file_path: str,
                      trajectory_file_path: str,
                      ckpt_file_path: str | None = None,
                      initial_magnetic_moments: list[float] | None = None,
                      relax_intervals: list[list[float]] | None = None,
                      ecut: float = 350,
                      xc: str = 'PBE',
                      nbands: int = -10,
                      mode_name: str = 'pw',
                      steps: int = 100,
                      force_threshold: float = FORCE_THRESHOLD):

    logfile = open(log_file_path, 'w')

    # check if the file at ckpt_file_path exists. If exists, resume the calculation from the ckpt_file_path
    if ckpt_file_path and os.path.exists(ckpt_file_path):
        atoms = resume_relax_calculation(ckpt_file_path=ckpt_file_path,
                                         trajectory_file_path=trajectory_file_path,
                                         force_threshold=force_threshold,
                                         steps=steps)
        return atoms

    if not is_bulk:
        kpoints[2] = 1
    # print(atoms.get_chemical_symbols())
    chemical_symbols_list = list(set(atoms.get_chemical_symbols()))
    chemical_symbols_list.sort()
    if isinstance(initial_magnetic_moments, list) and len(initial_magnetic_moments) == len(atoms):
        atoms.set_initial_magnetic_moments(initial_magnetic_moments)
    # set the initial magnetic moments by atom
    elif isinstance(initial_magnetic_moments, list) and len(initial_magnetic_moments) == len(chemical_symbols_list):
        initial_magnetic_moments_dict = {
            symbol: moment for symbol, moment in zip(chemical_symbols_list, initial_magnetic_moments)}
        initial_magnetic_moments_full_list = []
        for chemical_symbol in atoms.get_chemical_symbols():
            initial_magnetic_moments_full_list.append(
                initial_magnetic_moments_dict[chemical_symbol])
        atoms.set_initial_magnetic_moments(initial_magnetic_moments_full_list)
    else:
        atoms.set_initial_magnetic_moments([0] * len(atoms))
    # set the calculator
    calc = GPAW(mode={'name': mode_name, 'ecut': ecut},  # plane wave basis set cutoff
                xc=xc,  # EX-Crl functional (e.g., 'LDA', 'PBE', 'RPBE')
                nbands=nbands,  # total number of bands
                kpts={'size': kpoints, 'gamma': True},  # k-points
                spinpol=spinpol,  # non-spin-polarized calculation
                txt=status_file_path)  # output file

    atoms.calc = calc
    atoms = EnhancedAtoms(atoms)
    atoms.set_relax_region(relax_intervals)

    optimizer = BFGS(
        atoms=atoms, trajectory=trajectory_file_path, logfile=logfile)
    if ckpt_file_path:
        optimizer.attach(lambda: calc.write(ckpt_file_path), interval=1)
    # fmax is the convergence threshold for the force,  eV/A.
    optimizer.run(fmax=force_threshold, steps=steps)
    return atoms


def ground_state_calculation(atoms: Atoms,
                             kpoints: list[int],
                             spinpol: bool,
                             status_file_path: str,
                             ckpt_file_path: str | None = None,
                             initial_magnetic_moments: list[float] | None = None,
                             ecut: float = 350,
                             xc: str = 'PBE',
                             nbands: int = -10,
                             mode_name: str = 'pw',
                             n_write: int = 3):
    if ckpt_file_path and os.path.exists(ckpt_file_path):
        atoms = resume_ground_state_calculation(
            ckpt_file_path=ckpt_file_path, n_write=n_write)
        return atoms
    chemical_symbols_list = list(set(atoms.get_chemical_symbols()))
    chemical_symbols_list.sort()
    if isinstance(initial_magnetic_moments, list) and len(initial_magnetic_moments) == len(atoms):
        atoms.set_initial_magnetic_moments(initial_magnetic_moments)
    elif isinstance(initial_magnetic_moments, list) and len(initial_magnetic_moments) == len(chemical_symbols_list):
        initial_magnetic_moments_dict = {
            symbol: moment for symbol, moment in zip(
                chemical_symbols_list, initial_magnetic_moments)}
        atoms.set_initial_magnetic_moments_by_atom(
            initial_magnetic_moments_dict)
    else:
        atoms.set_initial_magnetic_moments([0] * len(atoms))
    calc = GPAW(mode={'name': mode_name, 'ecut': ecut},
                xc=xc,  # EX-Crl functional (e.g., 'LDA', 'PBE', 'RPBE')
                nbands=nbands,  # total number of bands
                kpts={'size': kpoints, 'gamma': True},
                spinpol=spinpol,
                txt=status_file_path)
    if ckpt_file_path:
        calc.attach(calc.write, n_write, ckpt_file_path, mode='all')
    atoms.calc = calc
    # carry the relaxation calculation
    e = atoms.get_potential_energy()

    print(f'Ground state (SCF) calculations finished. The energy is {e} eV')
    return atoms


def resume_ground_state_calculation(ckpt_file_path: str, n_write: int = 3):
    """Resume calculation from a checkpoint file."""
    if os.path.exists(ckpt_file_path):
        # Restart from checkpoint
        atoms, calc = restart(ckpt_file_path)
        if calc.scf.converged:
            atoms.calc = calc
            # The checkpoint is a converged calculation. Output results
            return atoms
        else:
            # The checkpoint is NOT converged. Resume calculation now.
            print(
                "The previous calculation is not converged. Resume the calculation now.")
            calc.attach(calc.write, n_write, ckpt_file_path, mode='all')
            atoms.get_potential_energy()
            return atoms


def band_calculation(atoms: Atoms,
                     status_file_path: str,
                     kpath: str | None = None,
                     npoints: int = 60,
                     symmetry: str = 'off',
                     include_soc: bool = True,
                     ckpt_file_path: str | None = None):
    # if status_file_path is not .txt, add .txt to the end of status_file_path
    # get the band path
    atoms = EnhancedAtoms(atoms)
    path = atoms.get_band_path(kpath=kpath, npoints=npoints)
    # to do: check if atoms.calc is defined and scf calculation is done
    calc = atoms.calc

    calc = calc.fixed_density(
        kpts=path,
        symmetry=symmetry,
        txt=status_file_path)

    atoms.calc = calc
    if ckpt_file_path:
        calc.write(ckpt_file_path, mode='all')
    return atoms


def perform_calculation_service(input_info: CalculationInfo,
                                output_info: CalculationInfo):
    """
    Perform the calculation for the given calculation type and parameters, based on the input information.
    This function assumes the input_info and output_info folders have been created.

    Args:
    - input_info: a CalculationInfo object containing the input data.
    - output_info: a CalculationInfo object containing the output data.

    Returns:
    - output_info: The output data after the calculation is performed.
    """
    atoms = input_info.get_atoms()
    calculation_type = output_info.calculation_type
    calculation_parameters = output_info.calculation_parameters
    # print(f"Calculation type: {calculation_type}")
    # print(f"Calculation parameters: {json.dumps(calculation_parameters, indent=4)}")
    output_info.update_status('processing')
    if calculation_type == 'relax':
        atoms = relax_calculation(atoms=atoms,
                                  status_file_path=output_info.status_file_path,
                                  ckpt_file_path=output_info.ckpt_file_path,
                                  trajectory_file_path=output_info.trajectory_file_path,
                                  log_file_path=output_info.log_file_path,
                                  **calculation_parameters)
    elif calculation_type == 'ground_state':
        atoms = ground_state_calculation(atoms=atoms,
                                         status_file_path=output_info.status_file_path,
                                         ckpt_file_path=output_info.ckpt_file_path,
                                         **calculation_parameters)
    elif calculation_type == 'band':
        # for band calculation, need to specify a ckpt file path to provide the initial charge density
        input_ckpt_file_path = getattr(input_info, 'ckpt_file_path', None)
        if not input_ckpt_file_path or not os.path.exists(input_ckpt_file_path):
            raise ValueError(
                "Input ckpt_file_path is empty. A ckpt file with charge density isrequired for band calculation. Maybe run relax or ground_state calculation first.")
        else:
            calc = GPAW(input_ckpt_file_path, txt=None)
            if not hasattr(calc, 'density') or calc.density is None:
                raise ValueError(
                    "Input ckpt_file_path is not a valid input because it does not have charge density.")
            atoms.calc = calc
            atoms = band_calculation(atoms=atoms,
                                     status_file_path=output_info.status_file_path,
                                     ckpt_file_path=output_info.ckpt_file_path,
                                     **calculation_parameters)
    # save the structure information in .cif file in the output info folder
    output_info.update_structure_from_atoms(atoms)
    output_info.update_status('finished')
    return output_info


def remove_calculation_subfolder(calculation_id: str, project_folder_path: str):
    """Remove the calculation subfolder in project_folder_path"""
    try:
        shutil.rmtree(os.path.join(project_folder_path, calculation_id))
    except FileNotFoundError:
        print(
            f"Subfolder {calculation_id} does not exist in {project_folder_path}.")


def perform_calculation_by_id(input_calculation_id: str, output_calculation_id: str, project_folder_path: str):
    """
    Perform the calculation based on the input information secified by input_calculation_id and output the result to the folder specified by output_information_id
    """
    input_info_file_path = os.path.join(
        project_folder_path, input_calculation_id, 'info.json')
    output_info_file_path = os.path.join(
        project_folder_path, output_calculation_id, 'info.json')
    input_info = CalculationInfo.from_file(input_info_file_path)
    output_info = CalculationInfo.from_file(output_info_file_path)
    # print(output_info)
    if not input_info:
        raise ValueError(
            f"No input information found in path {input_info_file_path}.")
    if not output_info:
        raise ValueError(
            f"No output information found in path {output_info_file_path}.")
    output_info = perform_calculation_service(input_info, output_info)
    return output_info


def perform_calculation(calculation_type: str,
                        calculation_parameters: dict,
                        project_folder_path: str,
                        output_calculation_id: str | None = None) -> str:
    """
    Start a relax calculation subprocessbased on the calculation type and parameters, and store the calculation info into a new calculation record.
    The input_calculation_id is chosen automatically. The output_calculation_id is a new uuid generated. If the calculation cannot be calculated,
    the function will return an error message.

    Args:
        relax_calculation_parameters: a dictionary containing the calculation parameters
        project_folder_path: the path to the project folder

    Returns:
        A dictionary containing the calculation id.
    """
    calculation_record = CalculationProject(
        project_folder_path=project_folder_path)
    if calculation_type == 'relax' or calculation_type == 'ground_state':
        input_info = calculation_record.latest_finished_calc()
    else:
        input_info = calculation_record.latest_dos_calc()
    if not input_info:
        input_info = calculation_record.info_list[0]
    if not output_calculation_id:
        output_calculation_id = str(uuid.uuid4())
    calculation_record.add_new(
        calculation_id=output_calculation_id,
        calculation_type=calculation_type,
        calculation_parameters=calculation_parameters)
    perform_calculation_by_id(input_calculation_id=input_info.calculation_id,
                              output_calculation_id=output_calculation_id,
                              project_folder_path=project_folder_path)
    return {
        "project_folder_path": project_folder_path,
        "calculation_id": output_calculation_id
    }


def check_calculation_result(calculation_id: str, project_folder_path: str):
    """
    Check the calculation result with the given calculation id and project folder path.

    Args:
        calculation_id: the id of the calculation
        project_folder_path: the path to the project folder

    Returns:
        A dictionary containing the calculation log.
    """
    calculation_info_file_path = os.path.join(
        project_folder_path, calculation_id, 'info.json')
    status_txt_file_path = os.path.join(
        project_folder_path, calculation_id, 'status.txt')
    calculation_info = None

    running = False
    for proc in psutil.process_iter(['pid', 'name']):
        if "python" not in proc.info['name']:
            continue
        for part in proc.cmdline():
            if "/main_run_calculation.py" in part and calculation_id in part:
                print(proc.cmdline())
                running = True
                break
        if running:
            break

    if not os.path.exists(calculation_info_file_path):
        return {
            "calculation_status": "failed",
            "log": "unable to start calculation"
        }
    elif not os.path.exists(status_txt_file_path) and not running:
        return {
            "calculation_status": "failed",
            "log": "the calculation might not be running"
        }

    start_time = time.time()
    while True:
        try:
            calculation_info = CalculationInfo.from_file(
                calculation_info_file_path)
            calculation_status = calculation_info.calculation_status
            status_file_content = calculation_info.status_file_content
        except Exception as e:
            status_file_content = str(e)
            calculation_status = "status inaccessible"
            print(f"Warning: {str(e)}", file=sys.stderr, flush=True)

        # result = {
        #         "calculation_status": calculation_status,
        #         "log": status_file_content
        #     }

        if calculation_status == 'finished':
            result = {
                "calculation_status": calculation_status,
                "log": status_file_content
            }
            # return the figure as a dictionary for JSON serialization
            try:
                fig = calculation_info.plot_output()
                if fig:
                    fig_str = pio.to_json(fig)
                    result["figure_str"] = fig_str
            except Exception as e:
                print(
                    f"Warning: Error generating plot: {str(e)}", file=sys.stderr, flush=True)
            break

        # if time exceeds max_computation_time, exit
        if time.time() - start_time > MAX_COMPUTATION_TIME:
            log = status_file_content + \
                f"\n\nCalculation exceeded maximum computation time {MAX_COMPUTATION_TIME} seconds"
            result = {
                "calculation_status": "timeout",
                "log": log
            }
            break

        time.sleep(5)

    return result


def stream_calculation_log(calculation_id: str, project_folder_path: str):
    """
    Stream the calculation log of a calculation subprocess, if the calculation is in progress.

    Args:
        calculation_id: the id of the calculation
        project_folder_path: the path to the project folder

    Returns:
        A dictionary containing the calculation log.
    """
    calculation_info_file_path = os.path.join(
        project_folder_path, calculation_id, 'info.json')

    while True:
        try:
            calculation_info = CalculationInfo.from_file(
                calculation_info_file_path)
            calculation_status = calculation_info.calculation_status
            status_file_content = calculation_info.status_file_content
            yield status_file_content
        except Exception as e:
            status_file_content = e
            calculation_status = "status inaccessible"

        if calculation_status == 'finished' or calculation_status == 'interrupted':
            break
        time.sleep(1)



import os
from typing import Annotated, List

import httpx
from mcp.types import TextContent, ImageContent
from pydantic import AnyUrl, Field
from mcp.server import FastMCP
import requests
import base64


async def fetch(url: str, jina_api_key: str) -> List[TextContent]:
    """Fetch a URL and return the content.

    Args:
        url: The URL to fetch
        jina_api_key: Jina API key

    Returns:
        TextContent containing canvas information
    """
    llm_results = []

    def crawl(url: str) -> str:
        headers = {"Content-Type": "application/json", "X-Return-Format": "markdown", "X-With-Images-Summary": "true"}
        if jina_api_key:
            headers["Authorization"] = f"Bearer {jina_api_key}"

        data = {"url": url}
        response = requests.post("https://r.jina.ai/", headers=headers, json=data)
        return response.text

    try:
        content = crawl(url)
        llm_results.append(TextContent(type="text", text=content))
    except Exception as e:
        llm_results.append(TextContent(type="text", text=f"Failed to fetch URL {url!r}: {e}"))

    return llm_results


mcp = FastMCP("mcp-web-fetch")


@mcp.tool(
    name="fetch-web",
    description="Fetch a URL and return the content. Images will be returned in ![]() format. DO NOT FETCH image_urls, for images use read-image-url instead.",
)
async def fetch_web(
    url: Annotated[AnyUrl, Field(description="URL to fetch")]
):
    return await fetch(str(url), os.getenv("JINA_API_KEY", ""))


@mcp.tool(
    name="read-image-url",
    description="Read images from URLs, convert to LLM readable format.",
)
def read_image_url(image_urls: List[str]) -> List[ImageContent| TextContent]:
    """Read images from URLs, convert to base64, and return the image content.

    Args:
        image_urls: List of URLs of the images to read

    Returns:
        ImageContent containing the base64 encoded image data
    """
    llm_results: list[ImageContent | TextContent] = []
    message_results: list[TextContent] = []

    for image_url in image_urls:
        # Convert to base64
        try:
            image_content = httpx.get(image_url).content
        except Exception as e:
            llm_results.append(TextContent(type="text", text=f"Failed to read image {image_url!r}: {e}"))
            message_results.append(TextContent(type="text", text=f"Failed to read image {image_url!r}: {e}"))
            continue
        mime_type = "image/jpeg"
        if image_content.startswith(b"\xff\xd8\xff"):
            mime_type = "image/jpeg"
        elif image_content.startswith(b"\x89PNG\r\n\x1a\n"):
            mime_type = "image/png"
        elif image_content.startswith(b"RIFF") and b"WEBP" in image_content[:12]:
            mime_type = "image/webp"
        else:
            llm_results.append(
                TextContent(
                    type="text",
                    text=f"Unsupported image type {mime_type!r} for {image_url!r}: only jpeg, png, webp are supported",
                )
            )
            message_results.append(
                TextContent(
                    type="text",
                    text=f"Unsupported image type {mime_type!r} for {image_url!r}: , only jpeg, png, webp are supported",
                )
            )
            continue
        image_data = base64.standard_b64encode(image_content).decode("utf-8")
        llm_results.append(ImageContent(type="image", data=image_data, mimeType=mime_type))

    return llm_results


import paramiko
from getpass import getpass
import os
from loguru import logger
from .config import load_ssh_config


def connect_ssh(hostname, username):
    ssh = paramiko.SSHClient()
    ssh.load_system_host_keys()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(hostname, username=username)
    logger.info(f"Successfully connected to {hostname}")
    return ssh


def connect_ssh_by_key(hostname, username, key_path):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        # Expand the key_path if it uses ~
        key_path = os.path.expanduser(key_path)

        # Check if the key file exists
        if not os.path.exists(key_path):
            raise FileNotFoundError(f"SSH key file not found: {key_path}")

        # Try to load the key
        try:
            key = paramiko.RSAKey.from_private_key_file(key_path)
        except paramiko.ssh_exception.PasswordRequiredException:
            # If a password is required, prompt the user
            passphrase = getpass(
                "Enter passphrase for key '{}': ".format(key_path))
            key = paramiko.RSAKey.from_private_key_file(
                key_path, password=passphrase)

        # Connect with the key
        ssh.connect(hostname, username=username, pkey=key)
        logger.info(f"Successfully connected to {hostname}")
        return ssh
    except Exception as e:
        raise e


def execute_command_on_server(command: str):
    """
    SSH to a computation server and execute a command.

    Args:
        command: The command to execute.
        ssh_config: The configuration of the SSH client. ssh_config needs to contain
            the hostname, username, key_path that are used to connect to the server. In
            addition, it contains remote_root_path, which is the absolute path of the
            parent folder at the remote server where the command will be executed

            Default to the config['ssh_config'] if not specified.
    """
    ssh_config = load_ssh_config()
    hostname = ssh_config["hostname"]
    username = ssh_config["username"]

    # use_key_to_connect is a flag to determine whether to use SSH key to connect
    # If use_key_to_connect is not specified in secrets.json, default to False
    use_key_to_connect = ssh_config.get("use_key_to_connect", False)
    logger.info(f"Connecting to {hostname} as {username}")
    print(f"Connecting to {hostname} as {username}")
    if use_key_to_connect:
        ssh = connect_ssh_by_key(
            hostname, username, key_path=ssh_config["key_filename"])
    else:
        ssh = connect_ssh(hostname, username)

    logger.info(f"Connected {ssh}")

    # remote_folder_path is the repo root path of the remote folder
    remote_root_path = ssh_config["remote_root_path"]

    ssh_command = f"cd {remote_root_path} && {command}"
    logger.info(f"Sending SSH command: \n\n{ssh_command}")
    stdin, stdout, stderr = ssh.exec_command(ssh_command, timeout=60 * 10)
    logger.info("executed")
    return stdin, stdout, stderr



# ssh to server and run gpaw calculation. Send log and calculation status back to local. When finished, return figure.
import base64
import json
import os
import shlex
import uuid
from typing import Literal, Optional, TypedDict, cast, Annotated

import plotly.io as pio
from loguru import logger
from mcp.server.fastmcp import FastMCP
from mcp.types import EmbeddedResource, ImageContent, TextContent, TextResourceContents

from .structure_helper import get_structure_info
from .ssh_utils import execute_command_on_server
from .config import load_config

mcp = FastMCP("dft_computation")
config = load_config()['DEFAULT']

remote_calculation_output_abs_path = config["remote_calculation_output_abs_path"]
remote_python_path = config["remote_python_path"]
remote_gpaw_path = config["remote_gpaw_path"]


def extract_project_id_from_uri(project_uri: str) -> str:
    """Extract the project ID from a project URI.

    Args:
        project_uri: The project URI to extract the project ID from

    Returns:
        bool: True if the URI has the correct format and contains a valid UUID
    """
    import uuid

    if not project_uri.startswith("project://"):
        raise ValueError("Project URI must start with 'project://'")

    try:
        # Extract the UUID part and validate it
        uuid_str = project_uri.split("://")[1]
        uuid.UUID(uuid_str)
        # Ensure string representation matches original (prevents invalid formats that uuid.UUID corrects)
        return uuid_str
    except (ValueError, IndexError):
        raise ValueError("Invalid UUID in project URI")


class RelaxCalculationParams(TypedDict):
    """Parameters for a gpaw relax calculation.

    Attributes:
        calculation_type: must be 'relax'
        spinpol: The boolean parameter determines whether the calculation is spin polarized.
        initial_magnetic_moments: An optional list that sets the initial magnetic moments for the atoms. This parameter is only needed if spinpol is True.
        kpoints: A list of 3 positive integers, representing the number of k-points in each direction of the reciprocal space. For example [1,1,1] or [4,4,2]
        is_bulk: The boolean parameter determines whether the calculation is for a bulk structure (periodic in all three directions) or a slab structure (periodic in two directions and open boundary condition in the third direction
        relax_intervals: A list of lists, such as [[z_1,z_2]] or [[z_1,z_2],[z_3,z_4]] which specifies intervals of z coordinates (in unit of angstrom) for relaxing atoms. If not selected, all atoms will be relaxed.
    """

    calculation_type: Literal["relax"]
    spinpol: bool
    initial_magnetic_moments: Optional[list[float]]
    kpoints: list[int]
    is_bulk: bool
    relax_intervals: Optional[list[list[float]]]


class GroundStateCalculationParams(TypedDict):
    """Parameters for a gpaw ground state calculation.

    Attributes:
        calculation_type: must be 'ground_state'
        spinpol: The boolean parameter determines whether the calculation is spin polarized.
        initial_magnetic_moments: A list that sets the initial magnetic moments for the atoms. This parameter is only needed if spinpol is True.
        kpoints: A list of 3 positive integers, representing the number of k-points in each direction of the reciprocal space. For example [1,1,1] or [4,4,2]
    """

    calculation_type: Literal["ground_state"]
    spinpol: bool
    initial_magnetic_moments: Optional[list[float]]
    kpoints: list[int]


class BandCalculationParams(TypedDict):
    """Parameters for a gpaw band structure calculation.

    Attributes:
        calculation_type: must be 'band'
        kpath: The string containing high-symmetry points that defines the k-path. For example 'GXMYG' or 'GXMYG,XG' if there are multiple segments. Use 'G' to represent the gamma point.
        npoints: The number of k-points along each segment of the k-path.
        include_soc: The parameter determines whether the calculation includes spin-orbit coupling.
    """

    calculation_type: Literal["band"]
    kpath: str
    npoints: int
    include_soc: Optional[bool]


@mcp.tool(name="create_new_project", description="Create a new project based on a given structure.")
async def create_new_project(
    structure_uri: Annotated[str, "structure_uri"],
    structure_type: Annotated[Literal["bulk", "supercell"], "structure_type, bulk or supercell"],
) -> TextContent:
    """
    Create a new project based on a given structure.

    Args:
        structure_uri: the uri of the structure with the format 'structure://{structure_id}
        structure_type: bulk (bulk crystal structure) or supercell (a thin film or supercell structure)

    Returns:
        The TextContext that contains the project id and a list of calculation info. The
        list of calculation info should contain only one item that is the initial structure information.
    """
    structure_info = get_structure_info(structure_uri)
    if structure_info is None:
        response = TextContent(type="text", text="Error:Structure not found")
        return response
    if structure_type == "bulk":
        structure = structure_info.structure
    elif structure_type == "supercell":
        structure = structure_info.structure
        if not structure:
            response = TextContent(
                type="text", text="Error: Supercell structure is not found.")
            return response
    structure_str = structure.to(fmt="json")

    project_id = str(uuid.uuid4())

    project_folder_path = os.path.join(
        remote_calculation_output_abs_path, project_id)

    structure_str = shlex.quote(structure_str)
    shell_command = (f"{remote_python_path} -m "
                     f"gpaw_computation_server.create_calculation_project "
                     f"-p {project_folder_path} -s {structure_str}")
    logger.info(f"shell_command: {str(shell_command)}")
    _, stdout, stderr = execute_command_on_server(shell_command)
    stdout_content = stdout.read().decode("utf-8")
    logger.info(f"stdout_content: {str(stdout_content)}")
    stderr_content = stderr.read().decode("utf-8")
    logger.info(f"stderr_content: {str(stderr_content)}")
    result = json.loads(stdout_content)
    result["project_uri"] = "project://" + project_id
    return TextContent(type="text", text=json.dumps(result))


@mcp.tool(name="start_calculation", description="Start a DFT calculation.")
async def start_calculation(
    params: Annotated[RelaxCalculationParams | GroundStateCalculationParams | BandCalculationParams, "params"],
    project_uri: str
) -> TextContent:
    """
    SSH to a computation server and start a DFT computation. There are three types of computation,
    1. 'relax': Relaxation of the structure
    2. 'ground_state': Self-consistent calculation of the charge density
    3. 'band': Calculation of the electron band structure, which requires a pre-calculated charge density,
    so it can only be carried if at least one relax or ground state calculation has been finished.

    Args:
        params: Parameters for the computation. Note that different types of computation requires different parameter fields.
        project_uri: The uri of the project. A project is defined by a structure. When the structure being studied
            is changed, one needs to start a new project

    Returns:
        dict[str, str]: A dictionary with the keys 'project_folder_path' and 'calculation_id'.
            - project_folder_path: The path of the project folder on the computation server.
            - calculation_id: The id of the new calculation.
    """
    logger.info("start calculation...")
    try:
        project_id = extract_project_id_from_uri(project_uri)
    except ValueError as e:
        return TextContent(type="text", text=f"Error: {str(e)}")

    output_calculation_id = str(uuid.uuid4())
    params_str = json.dumps(params)
    project_folder_path = os.path.join(
        remote_calculation_output_abs_path, project_id)

    shell_command = (
        f"/usr/bin/mpirun -np 4 {remote_gpaw_path} "
        f"python -m gpaw_computation_server.run_calculation "
        f"-pm {shlex.quote(params_str)} "
        f"-p {project_folder_path} "
        f"-c {output_calculation_id}"
    )
    logger.info(f"Executing command: {shell_command}")
    _, _, stderr = execute_command_on_server(shell_command)
    stderr_content = stderr.read().decode("utf-8")

    result = {
        "calculation_id": output_calculation_id,
        "project_id": project_id,
    }
    if stderr_content:
        result["error_log"] = stderr_content
    logger.info(f"Result: {result}")
    return TextContent(type="text", text=json.dumps(result))


@mcp.tool(name="check_calculation_result", description="Check the result of a DFT computation.")
async def check_calculation_result(
    output_calculation_id: Annotated[str, "output_calculation_id"],
    project_uri: Annotated[str, "project_uri"]
) -> list[TextContent | ImageContent | EmbeddedResource]:
    """
    SSH to a computation server and check the result of a DFT computation.

    Args:
        output_calculation_id: The id of the calculation to be checked.
        project_uri: The uri of the project. A project is defined by a structure. Each
            project contains a list of calculations.

    Returns:
        list[TextContent | ImageContent]: The list that contains the calculation_status,
        log and possibly the output figure.
            - The TextContent object contains the calculation status and log.
            - If the calculation is finished, the response will contain a
            ImageContent object that contains the figure.
    """
    try:
        project_id = extract_project_id_from_uri(project_uri)
    except ValueError as e:
        return [TextContent(type="text", text=f"Error: {str(e)}")]

    project_folder_path = os.path.join(
        remote_calculation_output_abs_path, project_id)

    shell_command = (
        f"{remote_python_path} -m "
        "gpaw_computation_server.check_calculation_result "
        f"-p {project_folder_path} -c {output_calculation_id}"
    )
    logger.info(f"actual python command, {shell_command}")
    _, stdout, stderr = execute_command_on_server(shell_command)
    try:
        stdout_content = stdout.read().decode("utf-8")
        result = json.loads(stdout_content)
    except Exception as e:
        logger.info(f"Error: {e}")
        logger.info("Error message from server:\n" +
                    stderr.read().decode("utf-8"))
        return [TextContent(type="text", text=f"Error: {e}")]

    text_result = {
        "calculation_status": result["calculation_status"],
        "log": result["log"],
    }
    response: list[TextContent | ImageContent | EmbeddedResource] = [
        TextContent(type="text", text=json.dumps(text_result, indent=4))
    ]

    if result.get("figure_str"):
        # Convert plotly figure data to actual figure
        fig = pio.from_json(result["figure_str"])

        # Convert to PNG image bytes
        img_bytes = fig.to_image(format="png", scale=1)
        # Convert to base64 string
        img_base64 = base64.b64encode(img_bytes).decode("utf-8")
        response.append(ImageContent(
            type="image", data=img_base64, mimeType="image/png"))
        response.append(
            EmbeddedResource(
                type="resource",
                resource=TextResourceContents(
                    uri=project_uri, text=cast(str, fig.to_json()), mimeType="application/json"
                ),
                # EmbeddedResource has enabled extra fields, so add an extra_type to indicate it's a plotly figure
                extra_type="plotly",
            )
        )
    return response


if __name__ == "__main__":
    # Initialize and run the server
    logger.info("starting gpaw computation server...")
    mcp.run(transport="stdio")



from dotenv import load_dotenv
import argparse
import logging
import os


# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)

# Load environment variables from .env file
# Get the directory of the current file
dotenv_path = os.path.join(os.path.dirname(__file__), '.env')
load_dotenv(dotenv_path=dotenv_path)
logger.info(f"Loaded environment variables from {dotenv_path}")


def parse_arguments():
    """Parse command line arguments for the SSH execution MCP server.

    Returns:
        argparse.Namespace: Parsed command line arguments
    """
    parser = argparse.ArgumentParser(description="SSH execution MCP server")

    # SSH connection configuration
    ssh_group = parser.add_argument_group("SSH Connection Configuration")
    ssh_group.add_argument(
        "--ssh-host", "-sh",
        type=str,
        help="SSH host to connect to (overrides SSH_HOST environment variable)"
    )
    ssh_group.add_argument(
        "--ssh-port", "-sp",
        type=int,
        help="SSH port to connect to (overrides SSH_PORT environment variable)"
    )
    ssh_group.add_argument(
        "--ssh-username", "-su",
        type=str,
        help="SSH username (overrides SSH_USERNAME environment variable)"
    )

    # Security configuration arguments
    security_group = parser.add_argument_group("Security Configuration")
    security_group.add_argument(
        "--allowed-commands", "-ac",
        type=str,
        help="Comma-separated list of commands that are allowed to be executed"
    )
    security_group.add_argument(
        "--allowed-paths", "-ap",
        type=str,
        help="Comma-separated list of paths that are allowed for command execution"
    )
    security_group.add_argument(
        "--commands-blacklist", "-cb",
        type=str,
        default="rm,mv,dd,mkfs,fdisk,format",
        help="Comma-separated list of commands that are not allowed (default: rm,mv,dd,mkfs,fdisk,format)"
    )
    security_group.add_argument(
        "--arguments-blacklist", "-ab",
        type=str,
        default="-rf,-fr,--force",
        help="Comma-separated list of arguments that are not allowed "
             "(default: -rf,-fr,--force)"
    )

    return parser.parse_args()


def update_environment_from_args(args):
    """Update environment variables from command-line arguments.

    Args:
        args (argparse.Namespace): Parsed command line arguments
    """
    env_mappings = {
        "ssh_host": "SSH_HOST",
        "ssh_port": "SSH_PORT",
        "ssh_username": "SSH_USERNAME",
        "allowed_commands": "SSH_ALLOWED_COMMANDS",
        "allowed_paths": "SSH_ALLOWED_PATHS",
        "commands_blacklist": "SSH_COMMANDS_BLACKLIST",
        "arguments_blacklist": "SSH_ARGUMENTS_BLACKLIST",
    }

    for arg_name, env_var in env_mappings.items():
        arg_value = getattr(args, arg_name, None)
        if arg_value is not None:
            os.environ[env_var] = str(arg_value)


def log_configuration():
    """Log the current server configuration."""
    # Use get() to safely access environment variables that might not be set
    config_vars = {
        "SSH host": os.environ.get('SSH_HOST', 'Not set'),
        "SSH port": os.environ.get('SSH_PORT', 'Not set'),
        "SSH username": os.environ.get('SSH_USERNAME', 'Not set'),
        "Allowed commands": os.environ.get('SSH_ALLOWED_COMMANDS', ''),
        "Allowed paths": os.environ.get('SSH_ALLOWED_PATHS', ''),
        "Commands blacklist": os.environ.get('SSH_COMMANDS_BLACKLIST', ''),
        "Arguments blacklist": os.environ.get('SSH_ARGUMENTS_BLACKLIST', ''),
    }

    logger.info("Starting SSH execution MCP server")

    for name, value in config_vars.items():
        logger.info(f"{name}: {value}")


def main():
    """Entry point for the SSH execution MCP server."""
    # Parse command line arguments
    args = parse_arguments()

    # Update environment variables from command-line arguments
    update_environment_from_args(args)

    # Log server configuration
    log_configuration()

    # Start the MCP server using the centralized function from server.py
    from .server import main as server_main
    server_main(transport="stdio")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding=utf-8

# This file is modified from the original file in the HuggingFace Inc. team.
# https://huggingface.co/docs/smolagents/en/index
import ast
import builtins
import difflib
import inspect
import logging
import resource
import sys
import re
from collections.abc import Mapping
from functools import wraps
from importlib import import_module
from types import BuiltinFunctionType, FunctionType, ModuleType
from typing import Any, Callable, Dict, List, Optional, Set
from .schemas import MAX_LENGTH_TRUNCATE_CONTENT, MAX_OPERATIONS, MAX_WHILE_ITERATIONS, BASE_BUILTIN_MODULES, DEFAULT_MAX_LEN_OUTPUT, DANGEROUS_FUNCTIONS, BASE_PYTHON_TOOLS, send_image_to_client
from mcp.types import ImageContent, EmbeddedResource
logger = logging.getLogger(__name__)


class InterpreterError(ValueError):
    """
    An error raised when the interpreter cannot evaluate a Python expression, due to syntax error or unsupported
    operations.
    """

    pass


ERRORS = {
    name: getattr(builtins, name)
    for name in dir(builtins)
    if isinstance(getattr(builtins, name), type) and issubclass(getattr(builtins, name), BaseException)
}


def truncate_content(content: str, max_length: int = MAX_LENGTH_TRUNCATE_CONTENT) -> str:
    if len(content) <= max_length:
        return content
    else:
        return (
            content[: max_length // 2]
            + f"\n..._This content has been truncated to stay below {max_length} characters_...\n"
            + content[-max_length // 2:]
        )


class PrintContainer:
    def __init__(self):
        self.value = ""
        self.images: list[ImageContent] = []

    def append(self, text):
        self.value += text
        return self

    def __iadd__(self, other):
        """Implements the += operator"""
        self.value += str(other)
        return self

    def __str__(self):
        """String representation"""
        return self.value

    def __repr__(self):
        """Representation for debugging"""
        return f"PrintContainer({self.value})"

    def __len__(self):
        """Implements len() function support"""
        return len(self.value)


class BreakException(Exception):
    pass


class ContinueException(Exception):
    pass


class ReturnException(Exception):
    def __init__(self, value):
        self.value = value


def get_iterable(obj):
    if isinstance(obj, list):
        return obj
    elif hasattr(obj, "__iter__"):
        return list(obj)
    else:
        raise InterpreterError("Object is not iterable")


def fix_final_answer_code(code: str) -> str:
    """
    Sometimes an LLM can try to assign a variable to final_answer, which would break the final_answer() tool.
    This function fixes this behaviour by replacing variable assignments to final_answer with final_answer_variable,
    while preserving function calls to final_answer().
    """
    # First, find if there's a direct assignment to final_answer
    # Use word boundary and negative lookbehind to ensure it's not an object attribute
    assignment_pattern = r"(?<!\.)(?<!\w)\bfinal_answer\s*="
    if "final_answer(" not in code or not re.search(assignment_pattern, code):
        # If final_answer tool is not called in this blob, then doing the replacement is hazardous because it could false the model's memory for next steps.
        # Let's not modify the code and leave the subsequent assignment error happen.
        return code

    # Pattern for replacing variable assignments
    # Looks for 'final_answer' followed by '=' with optional whitespace
    # Negative lookbehind ensures we don't match object attributes
    assignment_regex = r"(?<!\.)(?<!\w)(\bfinal_answer)(\s*=)"
    code = re.sub(assignment_regex, r"final_answer_variable\2", code)

    # Pattern for replacing variable usage but not function calls
    # Negative lookahead (?!\s*\() ensures we don't match function calls
    # Negative lookbehind (?<!\.|\w) ensures we don't match object methods or other variables
    variable_regex = r"(?<!\.)(?<!\w)(\bfinal_answer\b)(?!\s*\()"
    code = re.sub(variable_regex, "final_answer_variable", code)
    return code


def safer_eval(func: Callable):
    """
    Decorator to make the evaluation of a function safer by checking its return value.

    Args:
        func: Function to make safer.

    Returns:
        Callable: Safer function with return value check.
    """

    @wraps(func)
    def _check_return(
        expression,
        state,
        static_tools,
        custom_tools,
        authorized_imports=BASE_BUILTIN_MODULES,
    ):
        result = func(expression, state, static_tools,
                      custom_tools, authorized_imports=authorized_imports)
        if "*" not in authorized_imports:
            if isinstance(result, ModuleType):
                if not check_module_authorized(result.__name__, authorized_imports):
                    raise InterpreterError(
                        f"Forbidden access to module: {result.__name__}")
            elif isinstance(result, dict) and result.get("__spec__"):
                if not check_module_authorized(result["__name__"], authorized_imports):
                    raise InterpreterError(
                        f"Forbidden access to module: {result['__name__']}")
            elif isinstance(result, (FunctionType, BuiltinFunctionType)):
                for qualified_function_name in DANGEROUS_FUNCTIONS:
                    module_name, function_name = qualified_function_name.rsplit(
                        ".", 1)
                    if (
                        function_name not in static_tools
                        and result.__name__ == function_name
                        and result.__module__ == module_name
                    ):
                        raise InterpreterError(
                            f"Forbidden access to function: {function_name}")
        return result

    return _check_return


def evaluate_attribute(
    expression: ast.Attribute,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    if expression.attr.startswith("__") and expression.attr.endswith("__"):
        raise InterpreterError(
            f"Forbidden access to dunder attribute: {expression.attr}")
    value = evaluate_ast(expression.value, state,
                         static_tools, custom_tools, authorized_imports)
    return getattr(value, expression.attr)


def evaluate_unaryop(
    expression: ast.UnaryOp,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    operand = evaluate_ast(expression.operand, state,
                           static_tools, custom_tools, authorized_imports)
    if isinstance(expression.op, ast.USub):
        return -operand
    elif isinstance(expression.op, ast.UAdd):
        return operand
    elif isinstance(expression.op, ast.Not):
        return not operand
    elif isinstance(expression.op, ast.Invert):
        return ~operand
    else:
        raise InterpreterError(
            f"Unary operation {expression.op.__class__.__name__} is not supported.")


def evaluate_lambda(
    lambda_expression: ast.Lambda,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Callable:
    args = [arg.arg for arg in lambda_expression.args.args]

    def lambda_func(*values: Any) -> Any:
        new_state = state.copy()
        for arg, value in zip(args, values):
            new_state[arg] = value
        return evaluate_ast(
            lambda_expression.body,
            new_state,
            static_tools,
            custom_tools,
            authorized_imports,
        )

    return lambda_func


def evaluate_while(
    while_loop: ast.While,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    iterations = 0
    while evaluate_ast(while_loop.test, state, static_tools, custom_tools, authorized_imports):
        for node in while_loop.body:
            try:
                evaluate_ast(node, state, static_tools,
                             custom_tools, authorized_imports)
            except BreakException:
                return None
            except ContinueException:
                break
        iterations += 1
        if iterations > MAX_WHILE_ITERATIONS:
            raise InterpreterError(
                f"Maximum number of {MAX_WHILE_ITERATIONS} iterations in While loop exceeded")
    return None


def create_function(
    func_def: ast.FunctionDef,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Callable:
    source_code = ast.unparse(func_def)

    def new_func(*args: Any, **kwargs: Any) -> Any:
        func_state = state.copy()
        arg_names = [arg.arg for arg in func_def.args.args]
        default_values = [
            evaluate_ast(d, state, static_tools, custom_tools, authorized_imports) for d in func_def.args.defaults
        ]

        # Apply default values
        defaults = dict(zip(arg_names[-len(default_values):], default_values))

        # Set positional arguments
        for name, value in zip(arg_names, args):
            func_state[name] = value

        # Set keyword arguments
        for name, value in kwargs.items():
            func_state[name] = value

        # Handle variable arguments
        if func_def.args.vararg:
            vararg_name = func_def.args.vararg.arg
            func_state[vararg_name] = args

        if func_def.args.kwarg:
            kwarg_name = func_def.args.kwarg.arg
            func_state[kwarg_name] = kwargs

        # Set default values for arguments that were not provided
        for name, value in defaults.items():
            if name not in func_state:
                func_state[name] = value

        # Update function state with self and __class__
        if func_def.args.args and func_def.args.args[0].arg == "self":
            if args:
                func_state["self"] = args[0]
                func_state["__class__"] = args[0].__class__

        result = None
        try:
            for stmt in func_def.body:
                result = evaluate_ast(
                    stmt, func_state, static_tools, custom_tools, authorized_imports)
        except ReturnException as e:
            result = e.value

        if func_def.name == "__init__":
            return None

        return result

    # Store original AST, source code, and name
    new_func.__ast__ = func_def
    new_func.__source__ = source_code
    new_func.__name__ = func_def.name

    return new_func


def evaluate_function_def(
    func_def: ast.FunctionDef,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Callable:
    custom_tools[func_def.name] = create_function(
        func_def, state, static_tools, custom_tools, authorized_imports)
    return custom_tools[func_def.name]


def evaluate_class_def(
    class_def: ast.ClassDef,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> type:
    class_name = class_def.name
    bases = [evaluate_ast(base, state, static_tools, custom_tools,
                          authorized_imports) for base in class_def.bases]
    class_dict = {}

    for stmt in class_def.body:
        if isinstance(stmt, ast.FunctionDef):
            class_dict[stmt.name] = evaluate_ast(
                stmt, state, static_tools, custom_tools, authorized_imports)
        elif isinstance(stmt, ast.Assign):
            for target in stmt.targets:
                if isinstance(target, ast.Name):
                    class_dict[target.id] = evaluate_ast(
                        stmt.value,
                        state,
                        static_tools,
                        custom_tools,
                        authorized_imports,
                    )
                elif isinstance(target, ast.Attribute):
                    class_dict[target.attr] = evaluate_ast(
                        stmt.value,
                        state,
                        static_tools,
                        custom_tools,
                        authorized_imports,
                    )
        else:
            raise InterpreterError(
                f"Unsupported statement in class body: {stmt.__class__.__name__}")

    new_class = type(class_name, tuple(bases), class_dict)
    state[class_name] = new_class
    return new_class


def evaluate_augassign(
    expression: ast.AugAssign,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    def get_current_value(target: ast.AST) -> Any:
        if isinstance(target, ast.Name):
            return state.get(target.id, 0)
        elif isinstance(target, ast.Subscript):
            obj = evaluate_ast(target.value, state, static_tools,
                               custom_tools, authorized_imports)
            key = evaluate_ast(target.slice, state, static_tools,
                               custom_tools, authorized_imports)
            return obj[key]
        elif isinstance(target, ast.Attribute):
            obj = evaluate_ast(target.value, state, static_tools,
                               custom_tools, authorized_imports)
            return getattr(obj, target.attr)
        elif isinstance(target, ast.Tuple):
            return tuple(get_current_value(elt) for elt in target.elts)
        elif isinstance(target, ast.List):
            return [get_current_value(elt) for elt in target.elts]
        else:
            raise InterpreterError(
                "AugAssign not supported for {type(target)} targets.")

    current_value = get_current_value(expression.target)
    value_to_add = evaluate_ast(
        expression.value, state, static_tools, custom_tools, authorized_imports)

    if isinstance(expression.op, ast.Add):
        if isinstance(current_value, list):
            if not isinstance(value_to_add, list):
                raise InterpreterError(
                    f"Cannot add non-list value {value_to_add} to a list.")
            current_value += value_to_add
        else:
            current_value += value_to_add
    elif isinstance(expression.op, ast.Sub):
        current_value -= value_to_add
    elif isinstance(expression.op, ast.Mult):
        current_value *= value_to_add
    elif isinstance(expression.op, ast.Div):
        current_value /= value_to_add
    elif isinstance(expression.op, ast.Mod):
        current_value %= value_to_add
    elif isinstance(expression.op, ast.Pow):
        current_value **= value_to_add
    elif isinstance(expression.op, ast.FloorDiv):
        current_value //= value_to_add
    elif isinstance(expression.op, ast.BitAnd):
        current_value &= value_to_add
    elif isinstance(expression.op, ast.BitOr):
        current_value |= value_to_add
    elif isinstance(expression.op, ast.BitXor):
        current_value ^= value_to_add
    elif isinstance(expression.op, ast.LShift):
        current_value <<= value_to_add
    elif isinstance(expression.op, ast.RShift):
        current_value >>= value_to_add
    else:
        raise InterpreterError(
            f"Operation {type(expression.op).__name__} is not supported.")

    # Update the state: current_value has been updated in-place
    set_value(
        expression.target,
        current_value,
        state,
        static_tools,
        custom_tools,
        authorized_imports,
    )

    return current_value


def evaluate_boolop(
    node: ast.BoolOp,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> bool:
    if isinstance(node.op, ast.And):
        for value in node.values:
            if not evaluate_ast(value, state, static_tools, custom_tools, authorized_imports):
                return False
        return True
    elif isinstance(node.op, ast.Or):
        for value in node.values:
            if evaluate_ast(value, state, static_tools, custom_tools, authorized_imports):
                return True
        return False


def evaluate_binop(
    binop: ast.BinOp,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    # Recursively evaluate the left and right operands
    left_val = evaluate_ast(binop.left, state, static_tools,
                            custom_tools, authorized_imports)
    right_val = evaluate_ast(
        binop.right, state, static_tools, custom_tools, authorized_imports)

    # Determine the operation based on the type of the operator in the BinOp
    if isinstance(binop.op, ast.Add):
        return left_val + right_val
    elif isinstance(binop.op, ast.Sub):
        return left_val - right_val
    elif isinstance(binop.op, ast.Mult):
        return left_val * right_val
    elif isinstance(binop.op, ast.Div):
        return left_val / right_val
    elif isinstance(binop.op, ast.Mod):
        return left_val % right_val
    elif isinstance(binop.op, ast.Pow):
        return left_val**right_val
    elif isinstance(binop.op, ast.FloorDiv):
        return left_val // right_val
    elif isinstance(binop.op, ast.BitAnd):
        return left_val & right_val
    elif isinstance(binop.op, ast.BitOr):
        return left_val | right_val
    elif isinstance(binop.op, ast.BitXor):
        return left_val ^ right_val
    elif isinstance(binop.op, ast.LShift):
        return left_val << right_val
    elif isinstance(binop.op, ast.RShift):
        return left_val >> right_val
    else:
        raise NotImplementedError(
            f"Binary operation {type(binop.op).__name__} is not implemented.")


def evaluate_assign(
    assign: ast.Assign,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    result = evaluate_ast(assign.value, state, static_tools,
                          custom_tools, authorized_imports)
    if len(assign.targets) == 1:
        target = assign.targets[0]
        set_value(target, result, state, static_tools,
                  custom_tools, authorized_imports)
    else:
        expanded_values = []
        for tgt in assign.targets:
            if isinstance(tgt, ast.Starred):
                expanded_values.extend(result)
            else:
                expanded_values.append(result)

        for tgt, val in zip(assign.targets, expanded_values):
            set_value(tgt, val, state, static_tools,
                      custom_tools, authorized_imports)
    return result


def set_value(
    target: ast.AST,
    value: Any,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    if isinstance(target, ast.Name):
        if target.id in static_tools:
            raise InterpreterError(
                f"Cannot assign to name '{target.id}': doing this would erase the existing tool!")
        state[target.id] = value
    elif isinstance(target, ast.Tuple):
        if not isinstance(value, tuple):
            if hasattr(value, "__iter__") and not isinstance(value, (str, bytes)):
                value = tuple(value)
            else:
                raise InterpreterError("Cannot unpack non-tuple value")
        if len(target.elts) != len(value):
            raise InterpreterError("Cannot unpack tuple of wrong size")
        for i, elem in enumerate(target.elts):
            set_value(elem, value[i], state, static_tools,
                      custom_tools, authorized_imports)
    elif isinstance(target, ast.Subscript):
        obj = evaluate_ast(target.value, state, static_tools,
                           custom_tools, authorized_imports)
        key = evaluate_ast(target.slice, state, static_tools,
                           custom_tools, authorized_imports)
        obj[key] = value
    elif isinstance(target, ast.Attribute):
        obj = evaluate_ast(target.value, state, static_tools,
                           custom_tools, authorized_imports)
        setattr(obj, target.attr, value)


def evaluate_call(
    call: ast.Call,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    if not isinstance(call.func, (ast.Call, ast.Lambda, ast.Attribute, ast.Name, ast.Subscript)):
        raise InterpreterError(
            f"This is not a correct function: {call.func}).")

    func, func_name = None, None

    if isinstance(call.func, ast.Call):
        func = evaluate_ast(call.func, state, static_tools,
                            custom_tools, authorized_imports)
    elif isinstance(call.func, ast.Lambda):
        func = evaluate_ast(call.func, state, static_tools,
                            custom_tools, authorized_imports)
    elif isinstance(call.func, ast.Attribute):
        obj = evaluate_ast(call.func.value, state, static_tools,
                           custom_tools, authorized_imports)
        func_name = call.func.attr
        if not hasattr(obj, func_name):
            raise InterpreterError(
                f"Object {obj} has no attribute {func_name}")
        func = getattr(obj, func_name)
    elif isinstance(call.func, ast.Name):
        func_name = call.func.id
        if func_name in state:
            func = state[func_name]
        elif func_name in static_tools:
            func = static_tools[func_name]
        elif func_name in custom_tools:
            func = custom_tools[func_name]
        elif func_name in ERRORS:
            func = ERRORS[func_name]
        else:
            raise InterpreterError(
                f"It is not permitted to evaluate other functions than the provided tools or functions defined/imported in previous code (tried to execute {call.func.id})."
            )
    elif isinstance(call.func, ast.Subscript):
        func = evaluate_ast(call.func, state, static_tools,
                            custom_tools, authorized_imports)
        if not callable(func):
            raise InterpreterError(
                f"This is not a correct function: {call.func}).")
        func_name = None

    args = []
    for arg in call.args:
        if isinstance(arg, ast.Starred):
            args.extend(evaluate_ast(arg.value, state, static_tools,
                        custom_tools, authorized_imports))
        else:
            args.append(evaluate_ast(arg, state, static_tools,
                        custom_tools, authorized_imports))

    kwargs = {
        keyword.arg: evaluate_ast(
            keyword.value, state, static_tools, custom_tools, authorized_imports)
        for keyword in call.keywords
    }

    if func_name == "super":
        if not args:
            if "__class__" in state and "self" in state:
                return super(state["__class__"], state["self"])
            else:
                raise InterpreterError("super() needs at least one argument")
        cls = args[0]
        if not isinstance(cls, type):
            raise InterpreterError("super() argument 1 must be type")
        if len(args) == 1:
            return super(cls)
        elif len(args) == 2:
            instance = args[1]
            return super(cls, instance)
        else:
            raise InterpreterError("super() takes at most 2 arguments")
    elif func_name == "print":
        state["_print_outputs"] += " ".join(map(str, args)) + "\n"
        return None
    elif func_name == "send_image_to_client":
        image_content: list[ImageContent |
                            EmbeddedResource] = send_image_to_client(args[0])
        state["_print_outputs"].images.extend(image_content)
        return None
    else:  # Assume it's a callable object
        if (inspect.getmodule(func) == builtins) and inspect.isbuiltin(func) and (func not in static_tools.values()):
            raise InterpreterError(
                f"Invoking a builtin function that has not been explicitly added as a tool is not allowed ({func_name})."
            )
        return func(*args, **kwargs)


def evaluate_subscript(
    subscript: ast.Subscript,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    index = evaluate_ast(subscript.slice, state, static_tools,
                         custom_tools, authorized_imports)
    value = evaluate_ast(subscript.value, state, static_tools,
                         custom_tools, authorized_imports)
    try:
        return value[index]
    except (KeyError, IndexError, TypeError) as e:
        error_message = f"Could not index {value} with '{index}': {type(e).__name__}: {e}"
        if isinstance(index, str) and isinstance(value, Mapping):
            close_matches = difflib.get_close_matches(
                index, list(value.keys()))
            if len(close_matches) > 0:
                error_message += f". Maybe you meant one of these indexes instead: {str(close_matches)}"
        raise InterpreterError(error_message) from e


def evaluate_name(
    name: ast.Name,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    if name.id in state:
        return state[name.id]
    elif name.id in static_tools:
        return static_tools[name.id]
    elif name.id in custom_tools:
        return custom_tools[name.id]
    elif name.id in ERRORS:
        return ERRORS[name.id]
    close_matches = difflib.get_close_matches(name.id, list(state.keys()))
    if len(close_matches) > 0:
        return state[close_matches[0]]
    raise InterpreterError(f"The variable `{name.id}` is not defined.")


def evaluate_condition(
    condition: ast.Compare,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> bool | object:
    result = True
    left = evaluate_ast(condition.left, state, static_tools,
                        custom_tools, authorized_imports)
    for i, (op, comparator) in enumerate(zip(condition.ops, condition.comparators)):
        op = type(op)
        right = evaluate_ast(comparator, state, static_tools,
                             custom_tools, authorized_imports)
        if op == ast.Eq:
            current_result = left == right
        elif op == ast.NotEq:
            current_result = left != right
        elif op == ast.Lt:
            current_result = left < right
        elif op == ast.LtE:
            current_result = left <= right
        elif op == ast.Gt:
            current_result = left > right
        elif op == ast.GtE:
            current_result = left >= right
        elif op == ast.Is:
            current_result = left is right
        elif op == ast.IsNot:
            current_result = left is not right
        elif op == ast.In:
            current_result = left in right
        elif op == ast.NotIn:
            current_result = left not in right
        else:
            raise InterpreterError(f"Unsupported comparison operator: {op}")

        if current_result is False:
            return False
        result = current_result if i == 0 else (result and current_result)
        left = right
    return result


def evaluate_if(
    if_statement: ast.If,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    result = None
    test_result = evaluate_ast(
        if_statement.test, state, static_tools, custom_tools, authorized_imports)
    if test_result:
        for line in if_statement.body:
            line_result = evaluate_ast(
                line, state, static_tools, custom_tools, authorized_imports)
            if line_result is not None:
                result = line_result
    else:
        for line in if_statement.orelse:
            line_result = evaluate_ast(
                line, state, static_tools, custom_tools, authorized_imports)
            if line_result is not None:
                result = line_result
    return result


def evaluate_for(
    for_loop: ast.For,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Any:
    result = None
    iterator = evaluate_ast(for_loop.iter, state,
                            static_tools, custom_tools, authorized_imports)
    for counter in iterator:
        set_value(
            for_loop.target,
            counter,
            state,
            static_tools,
            custom_tools,
            authorized_imports,
        )
        for node in for_loop.body:
            try:
                line_result = evaluate_ast(
                    node, state, static_tools, custom_tools, authorized_imports)
                if line_result is not None:
                    result = line_result
            except BreakException:
                break
            except ContinueException:
                continue
        else:
            continue
        break
    return result


def evaluate_listcomp(
    listcomp: ast.ListComp,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> List[Any]:
    def inner_evaluate(generators: List[ast.comprehension], index: int, current_state: Dict[str, Any]) -> List[Any]:
        if index >= len(generators):
            return [
                evaluate_ast(
                    listcomp.elt,
                    current_state,
                    static_tools,
                    custom_tools,
                    authorized_imports,
                )
            ]
        generator = generators[index]
        iter_value = evaluate_ast(
            generator.iter,
            current_state,
            static_tools,
            custom_tools,
            authorized_imports,
        )
        result = []
        for value in iter_value:
            new_state = current_state.copy()
            if isinstance(generator.target, ast.Tuple):
                for idx, elem in enumerate(generator.target.elts):
                    new_state[elem.id] = value[idx]
            else:
                new_state[generator.target.id] = value
            if all(
                evaluate_ast(if_clause, new_state, static_tools,
                             custom_tools, authorized_imports)
                for if_clause in generator.ifs
            ):
                result.extend(inner_evaluate(generators, index + 1, new_state))
        return result

    return inner_evaluate(listcomp.generators, 0, state)


def evaluate_setcomp(
    setcomp: ast.SetComp,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Set[Any]:
    result = set()
    for gen in setcomp.generators:
        iter_value = evaluate_ast(
            gen.iter, state, static_tools, custom_tools, authorized_imports)
        for value in iter_value:
            new_state = state.copy()
            set_value(
                gen.target,
                value,
                new_state,
                static_tools,
                custom_tools,
                authorized_imports,
            )
            if all(
                evaluate_ast(if_clause, new_state, static_tools,
                             custom_tools, authorized_imports)
                for if_clause in gen.ifs
            ):
                element = evaluate_ast(
                    setcomp.elt,
                    new_state,
                    static_tools,
                    custom_tools,
                    authorized_imports,
                )
                result.add(element)
    return result


def evaluate_try(
    try_node: ast.Try,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    try:
        for stmt in try_node.body:
            evaluate_ast(stmt, state, static_tools,
                         custom_tools, authorized_imports)
    except Exception as e:
        matched = False
        for handler in try_node.handlers:
            if handler.type is None or isinstance(
                e,
                evaluate_ast(handler.type, state, static_tools,
                             custom_tools, authorized_imports),
            ):
                matched = True
                if handler.name:
                    state[handler.name] = e
                for stmt in handler.body:
                    evaluate_ast(stmt, state, static_tools,
                                 custom_tools, authorized_imports)
                break
        if not matched:
            raise e
    else:
        if try_node.orelse:
            for stmt in try_node.orelse:
                evaluate_ast(stmt, state, static_tools,
                             custom_tools, authorized_imports)
    finally:
        if try_node.finalbody:
            for stmt in try_node.finalbody:
                evaluate_ast(stmt, state, static_tools,
                             custom_tools, authorized_imports)


def evaluate_raise(
    raise_node: ast.Raise,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    if raise_node.exc is not None:
        exc = evaluate_ast(raise_node.exc, state, static_tools,
                           custom_tools, authorized_imports)
    else:
        exc = None
    if raise_node.cause is not None:
        cause = evaluate_ast(raise_node.cause, state,
                             static_tools, custom_tools, authorized_imports)
    else:
        cause = None
    if exc is not None:
        if cause is not None:
            raise exc from cause
        else:
            raise exc
    else:
        raise InterpreterError(
            "Re-raise is not supported without an active exception")


def evaluate_assert(
    assert_node: ast.Assert,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    test_result = evaluate_ast(
        assert_node.test, state, static_tools, custom_tools, authorized_imports)
    if not test_result:
        if assert_node.msg:
            msg = evaluate_ast(assert_node.msg, state,
                               static_tools, custom_tools, authorized_imports)
            raise AssertionError(msg)
        else:
            # Include the failing condition in the assertion message
            test_code = ast.unparse(assert_node.test)
            raise AssertionError(f"Assertion failed: {test_code}")


def evaluate_with(
    with_node: ast.With,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    contexts = []
    for item in with_node.items:
        context_expr = evaluate_ast(
            item.context_expr, state, static_tools, custom_tools, authorized_imports)
        if item.optional_vars:
            state[item.optional_vars.id] = context_expr.__enter__()
            contexts.append(state[item.optional_vars.id])
        else:
            context_var = context_expr.__enter__()
            contexts.append(context_var)

    try:
        for stmt in with_node.body:
            evaluate_ast(stmt, state, static_tools,
                         custom_tools, authorized_imports)
    except Exception as e:
        for context in reversed(contexts):
            context.__exit__(type(e), e, e.__traceback__)
        raise
    else:
        for context in reversed(contexts):
            context.__exit__(None, None, None)


def get_safe_module(raw_module, authorized_imports, visited=None):
    """Creates a safe copy of a module or returns the original if it's a function"""
    # If it's a function or non-module object, return it directly
    if not isinstance(raw_module, ModuleType):
        return raw_module

    # Handle circular references: Initialize visited set for the first call
    if visited is None:
        visited = set()

    module_id = id(raw_module)
    if module_id in visited:
        return raw_module  # Return original for circular refs

    visited.add(module_id)

    # Create new module for actual modules
    safe_module = ModuleType(raw_module.__name__)

    # Copy all attributes by reference, recursively checking modules
    for attr_name in dir(raw_module):
        try:
            attr_value = getattr(raw_module, attr_name)
        except (ImportError, AttributeError) as e:
            # lazy / dynamic loading module -> INFO log and skip
            logger.info(
                f"Skipping import error while copying {raw_module.__name__}.{attr_name}: {type(e).__name__} - {e}"
            )
            continue
        # Recursively process nested modules, passing visited set
        if isinstance(attr_value, ModuleType):
            attr_value = get_safe_module(
                attr_value, authorized_imports, visited=visited)

        setattr(safe_module, attr_name, attr_value)

    return safe_module


def check_module_authorized(module_name, authorized_imports):
    if "*" in authorized_imports:
        return True
    else:
        # Check if the module is directly authorized
        if module_name in authorized_imports:
            return True

        # Check if any parent module is authorized (this allows submodules)
        module_path = module_name.split(".")
        # ["A", "B", "C"] -> ["A", "A.B", "A.B.C"]
        module_subpaths = [".".join(module_path[:i])
                           for i in range(1, len(module_path) + 1)]

        # Check if any parent module is in the authorized imports
        for auth_import in authorized_imports:
            # If the authorized import is a parent of the requested module
            if any(subpath == auth_import for subpath in module_subpaths):
                return True

        return False


def evaluate_import(expression, state, authorized_imports):
    if isinstance(expression, ast.Import):
        for alias in expression.names:
            if check_module_authorized(alias.name, authorized_imports):
                raw_module = import_module(alias.name)
                state[alias.asname or alias.name] = get_safe_module(
                    raw_module, authorized_imports)
            else:
                raise InterpreterError(
                    f"Import of {alias.name} is not allowed. Authorized imports are: {str(authorized_imports)}"
                )
        return None
    elif isinstance(expression, ast.ImportFrom):
        if check_module_authorized(expression.module, authorized_imports):
            raw_module = __import__(expression.module, fromlist=[
                                    alias.name for alias in expression.names])
            module = get_safe_module(raw_module, authorized_imports)
            if expression.names[0].name == "*":  # Handle "from module import *"
                if hasattr(module, "__all__"):  # If module has __all__, import only those names
                    for name in module.__all__:
                        state[name] = getattr(module, name)
                # If no __all__, import all public names (those not starting with '_')
                else:
                    for name in dir(module):
                        if not name.startswith("_"):
                            state[name] = getattr(module, name)
            else:  # regular from imports
                for alias in expression.names:
                    if hasattr(module, alias.name):
                        state[alias.asname or alias.name] = getattr(
                            module, alias.name)
                    else:
                        raise InterpreterError(
                            f"Module {expression.module} has no attribute {alias.name}")
        else:
            raise InterpreterError(
                f"Import from {expression.module} is not allowed. Authorized imports are: {str(authorized_imports)}"
            )
        return None


def evaluate_dictcomp(
    dictcomp: ast.DictComp,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> Dict[Any, Any]:
    result = {}
    for gen in dictcomp.generators:
        iter_value = evaluate_ast(
            gen.iter, state, static_tools, custom_tools, authorized_imports)
        for value in iter_value:
            new_state = state.copy()
            set_value(
                gen.target,
                value,
                new_state,
                static_tools,
                custom_tools,
                authorized_imports,
            )
            if all(
                evaluate_ast(if_clause, new_state, static_tools,
                             custom_tools, authorized_imports)
                for if_clause in gen.ifs
            ):
                key = evaluate_ast(
                    dictcomp.key,
                    new_state,
                    static_tools,
                    custom_tools,
                    authorized_imports,
                )
                val = evaluate_ast(
                    dictcomp.value,
                    new_state,
                    static_tools,
                    custom_tools,
                    authorized_imports,
                )
                result[key] = val
    return result


def evaluate_delete(
    delete_node: ast.Delete,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str],
) -> None:
    """
    Evaluate a delete statement (del x, del x[y]).

    Args:
        delete_node: The AST Delete node to evaluate
        state: The current state dictionary
        static_tools: Dictionary of static tools
        custom_tools: Dictionary of custom tools
        authorized_imports: List of authorized imports
    """
    for target in delete_node.targets:
        if isinstance(target, ast.Name):
            # Handle simple variable deletion (del x)
            if target.id in state:
                del state[target.id]
            else:
                raise InterpreterError(
                    f"Cannot delete name '{target.id}': name is not defined")
        elif isinstance(target, ast.Subscript):
            # Handle index/key deletion (del x[y])
            obj = evaluate_ast(target.value, state, static_tools,
                               custom_tools, authorized_imports)
            index = evaluate_ast(
                target.slice, state, static_tools, custom_tools, authorized_imports)
            try:
                del obj[index]
            except (TypeError, KeyError, IndexError) as e:
                raise InterpreterError(f"Cannot delete index/key: {str(e)}")
        else:
            raise InterpreterError(
                f"Deletion of {type(target).__name__} targets is not supported")


@safer_eval
def evaluate_ast(
    expression: ast.AST,
    state: Dict[str, Any],
    static_tools: Dict[str, Callable],
    custom_tools: Dict[str, Callable],
    authorized_imports: List[str] = BASE_BUILTIN_MODULES,
):
    """
    Evaluate an abstract syntax tree using the content of the variables stored in a state and only evaluating a given
    set of functions.

    This function will recurse through the nodes of the tree provided.

    Args:
        expression (`ast.AST`):
            The code to evaluate, as an abstract syntax tree.
        state (`Dict[str, Any]`):
            A dictionary mapping variable names to values. The `state` is updated if need be when the evaluation
            encounters assignments.
        static_tools (`Dict[str, Callable]`):
            Functions that may be called during the evaluation. Trying to change one of these static_tools will raise an error.
        custom_tools (`Dict[str, Callable]`):
            Functions that may be called during the evaluation. These static_tools can be overwritten.
        authorized_imports (`List[str]`):
            The list of modules that can be imported by the code. By default, only a few safe modules are allowed.
            If it contains "*", it will authorize any import. Use this at your own risk!
    """
    if state.setdefault("_operations_count", {"counter": 0})["counter"] >= MAX_OPERATIONS:
        raise InterpreterError(
            f"Reached the max number of operations of {MAX_OPERATIONS}. Maybe there is an infinite loop somewhere in the code, or you're just asking too many calculations."
        )
    state["_operations_count"]["counter"] += 1
    common_params = (state, static_tools, custom_tools, authorized_imports)
    if isinstance(expression, ast.Assign):
        # Assignment -> we evaluate the assignment which should update the state
        # We return the variable assigned as it may be used to determine the final result.
        return evaluate_assign(expression, *common_params)
    elif isinstance(expression, ast.AugAssign):
        return evaluate_augassign(expression, *common_params)
    elif isinstance(expression, ast.Call):
        # Function call -> we return the value of the function call
        return evaluate_call(expression, *common_params)
    elif isinstance(expression, ast.Constant):
        # Constant -> just return the value
        return expression.value
    elif isinstance(expression, ast.Tuple):
        return tuple((evaluate_ast(elt, *common_params) for elt in expression.elts))
    elif isinstance(expression, (ast.ListComp, ast.GeneratorExp)):
        return evaluate_listcomp(expression, *common_params)
    elif isinstance(expression, ast.DictComp):
        return evaluate_dictcomp(expression, *common_params)
    elif isinstance(expression, ast.SetComp):
        return evaluate_setcomp(expression, *common_params)
    elif isinstance(expression, ast.UnaryOp):
        return evaluate_unaryop(expression, *common_params)
    elif isinstance(expression, ast.Starred):
        return evaluate_ast(expression.value, *common_params)
    elif isinstance(expression, ast.BoolOp):
        # Boolean operation -> evaluate the operation
        return evaluate_boolop(expression, *common_params)
    elif isinstance(expression, ast.Break):
        raise BreakException()
    elif isinstance(expression, ast.Continue):
        raise ContinueException()
    elif isinstance(expression, ast.BinOp):
        # Binary operation -> execute operation
        return evaluate_binop(expression, *common_params)
    elif isinstance(expression, ast.Compare):
        # Comparison -> evaluate the comparison
        return evaluate_condition(expression, *common_params)
    elif isinstance(expression, ast.Lambda):
        return evaluate_lambda(expression, *common_params)
    elif isinstance(expression, ast.FunctionDef):
        return evaluate_function_def(expression, *common_params)
    elif isinstance(expression, ast.Dict):
        # Dict -> evaluate all keys and values
        keys = (evaluate_ast(k, *common_params) for k in expression.keys)
        values = (evaluate_ast(v, *common_params) for v in expression.values)
        return dict(zip(keys, values))
    elif isinstance(expression, ast.Expr):
        # Expression -> evaluate the content
        return evaluate_ast(expression.value, *common_params)
    elif isinstance(expression, ast.For):
        # For loop -> execute the loop
        return evaluate_for(expression, *common_params)
    elif isinstance(expression, ast.FormattedValue):
        # Formatted value (part of f-string) -> evaluate the content and format it
        value = evaluate_ast(expression.value, *common_params)
        # Early return if no format spec
        if not expression.format_spec:
            return value
        # Apply format specification
        format_spec = evaluate_ast(expression.format_spec, *common_params)
        return format(value, format_spec)
    elif isinstance(expression, ast.If):
        # If -> execute the right branch
        return evaluate_if(expression, *common_params)
    elif hasattr(ast, "Index") and isinstance(expression, ast.Index):
        return evaluate_ast(expression.value, *common_params)
    elif isinstance(expression, ast.JoinedStr):
        return "".join([str(evaluate_ast(v, *common_params)) for v in expression.values])
    elif isinstance(expression, ast.List):
        # List -> evaluate all elements
        return [evaluate_ast(elt, *common_params) for elt in expression.elts]
    elif isinstance(expression, ast.Name):
        # Name -> pick up the value in the state
        return evaluate_name(expression, *common_params)
    elif isinstance(expression, ast.Subscript):
        # Subscript -> return the value of the indexing
        return evaluate_subscript(expression, *common_params)
    elif isinstance(expression, ast.IfExp):
        test_val = evaluate_ast(expression.test, *common_params)
        if test_val:
            return evaluate_ast(expression.body, *common_params)
        else:
            return evaluate_ast(expression.orelse, *common_params)
    elif isinstance(expression, ast.Attribute):
        return evaluate_attribute(expression, *common_params)
    elif isinstance(expression, ast.Slice):
        return slice(
            evaluate_ast(
                expression.lower, *common_params) if expression.lower is not None else None,
            evaluate_ast(
                expression.upper, *common_params) if expression.upper is not None else None,
            evaluate_ast(
                expression.step, *common_params) if expression.step is not None else None,
        )
    elif isinstance(expression, ast.While):
        return evaluate_while(expression, *common_params)
    elif isinstance(expression, (ast.Import, ast.ImportFrom)):
        return evaluate_import(expression, state, authorized_imports)
    elif isinstance(expression, ast.ClassDef):
        return evaluate_class_def(expression, *common_params)
    elif isinstance(expression, ast.Try):
        return evaluate_try(expression, *common_params)
    elif isinstance(expression, ast.Raise):
        return evaluate_raise(expression, *common_params)
    elif isinstance(expression, ast.Assert):
        return evaluate_assert(expression, *common_params)
    elif isinstance(expression, ast.With):
        return evaluate_with(expression, *common_params)
    elif isinstance(expression, ast.Set):
        return set((evaluate_ast(elt, *common_params) for elt in expression.elts))
    elif isinstance(expression, ast.Return):
        raise ReturnException(evaluate_ast(
            expression.value, *common_params) if expression.value else None)
    elif isinstance(expression, ast.Pass):
        return None
    elif isinstance(expression, ast.Delete):
        return evaluate_delete(expression, *common_params)
    else:
        # For now we refuse anything else. Let's add things as we need them.
        raise InterpreterError(
            f"{expression.__class__.__name__} is not supported.")


def evaluate_python_code(
    code: str,
    static_tools: Optional[Dict[str, Callable]] = None,
    custom_tools: Optional[Dict[str, Callable]] = None,
    state: Optional[Dict[str, Any]] = None,
    authorized_imports: List[str] = BASE_BUILTIN_MODULES,
    max_print_outputs_length: int = DEFAULT_MAX_LEN_OUTPUT,
    # Resource limits
    max_memory_mb: int = 100,  # Maximum memory usage in MB
    max_cpu_time_sec: int = 15,  # Maximum CPU time in seconds
) -> tuple[str, list[ImageContent | EmbeddedResource]]:
    """
    Evaluate a python expression using the content of the variables stored in a state and only evaluating a given set
    of functions.

    This function will recurse through the nodes of the tree provided.

    Args:
        code (`str`):
            The code to evaluate.
        static_tools (`Dict[str, Callable]`):
            The functions that may be called during the evaluation. These can also be agents in a multiagent setting.
            These tools cannot be overwritten in the code: any assignment to their name will raise an error.
        custom_tools (`Dict[str, Callable]`):
            The functions that may be called during the evaluation.
            These tools can be overwritten in the code: any assignment to their name will overwrite them.
        state (`Dict[str, Any]`):
            A dictionary mapping variable names to values. The `state` should contain the initial inputs but will be
            updated by this function to contain all variables as they are evaluated.
            The print outputs will be stored in the state under the key "_print_outputs".

    Returns:
        str: Either the print outputs if successful, or print outputs + error message if failed
    """
    if state is None:
        state = {}
    static_tools = static_tools.copy() if static_tools is not None else {}

    # Add basic Python tools including print to static_tools
    for name, tool in BASE_PYTHON_TOOLS.items():
        if name not in static_tools:
            static_tools[name] = tool

    custom_tools = custom_tools if custom_tools is not None else {}
    state["_print_outputs"] = PrintContainer()
    state["_operations_count"] = {"counter": 0}

    # only for linux
    if sys.platform == "linux":
        resource.setrlimit(resource.RLIMIT_CPU,
                           (max_cpu_time_sec, max_cpu_time_sec))
        # Increase memory limit to 500MB to accommodate numpy
        if any(lib in authorized_imports for lib in ["numpy", "matplotlib", "plotly"]):
            max_memory_mb = 1000
        memory_limit = max_memory_mb * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_AS, (memory_limit *
                           1024 * 1024, memory_limit * 2 * 1024 * 1024))

    try:
        expression = ast.parse(code)
    except SyntaxError as e:
        error_msg = (
            f"Code parsing failed on line {e.lineno} due to: {type(e).__name__}\n"
            f"{e.text}"
            f"{' ' * (e.offset or 0)}^\n"
            f"Error: {str(e)}"
        )
        return error_msg, []

    try:
        # Use context manager to limit resources only during AST evaluation
        for node in expression.body:
            evaluate_ast(node, state, static_tools,
                         custom_tools, authorized_imports)

        return truncate_content(str(state["_print_outputs"]), max_length=max_print_outputs_length), state["_print_outputs"].images
    except (MemoryError, OSError, BlockingIOError) as e:
        # These exceptions are likely due to resource limits being hit
        current_output = str(state["_print_outputs"])
        resource_error_msg = (
            f"\n⚠️ RESOURCE LIMIT EXCEEDED ⚠️\n"
            f"This tool is meant for basic scientific calculations only.\n"
            f"Attempting to bypass resource limits or execute malicious code may result in account termination.\n"
            f"Error: {type(e).__name__}: {e}"
        )
        return truncate_content(current_output + resource_error_msg, max_length=max_print_outputs_length), state["_print_outputs"].images
    except Exception as e:
        current_output = str(state["_print_outputs"])
        error_msg = f"\nCode execution failed at line '{ast.get_source_segment(code, node)}' due to: {type(e).__name__}: {e}"
        return truncate_content(current_output + error_msg, max_length=max_print_outputs_length), state["_print_outputs"].images



from loguru import logger
from .server import mcp


def main():
    """Entry point for the Materials Project MCP server"""
    logger.info("starting materials project server...")
    mcp.run(transport="stdio")


if __name__ == "__main__":
    main()
